import numpy as np

from ase.atoms import Atoms


class Quaternions(Atoms):

    def __init__(self, *args, **kwargs):
        quaternions = None
        if 'quaternions' in kwargs:
            quaternions = np.array(kwargs['quaternions'])
            del kwargs['quaternions']
        Atoms.__init__(self, *args, **kwargs)
        if quaternions is not None:
            self.set_array('quaternions', quaternions, shape=(4,))
            # set default shapes
            self.set_shapes(np.array([[3, 2, 1]] * len(self)))

    def set_shapes(self, shapes):
        self.set_array('shapes', shapes, shape=(3,))

    def set_quaternions(self, quaternions):
        self.set_array('quaternions', quaternions, quaternion=(4,))

    def get_shapes(self):
        return self.get_array('shapes')

    def get_quaternions(self):
        return self.get_array('quaternions').copy()


class Quaternion:

    def __init__(self, qin=[1, 0, 0, 0]):
        assert len(qin) == 4
        self.q = np.array(qin)

    def __str__(self):
        return self.q.__str__()

    def __mul__(self, other):
        sw, sx, sy, sz = self.q
        ow, ox, oy, oz = other.q
        return Quaternion([sw * ow - sx * ox - sy * oy - sz * oz,
                           sw * ox + sx * ow + sy * oz - sz * oy,
                           sw * oy + sy * ow + sz * ox - sx * oz,
                           sw * oz + sz * ow + sx * oy - sy * ox])

    def conjugate(self):
        return Quaternion(-self.q * np.array([-1., 1., 1., 1.]))

    def rotate(self, vector):
        """Apply the rotation matrix to a vector."""
        qw, qx, qy, qz = self.q[0], self.q[1], self.q[2], self.q[3]
        x, y, z = vector[0], vector[1], vector[2]

        ww = qw * qw
        xx = qx * qx
        yy = qy * qy
        zz = qz * qz
        wx = qw * qx
        wy = qw * qy
        wz = qw * qz
        xy = qx * qy
        xz = qx * qz
        yz = qy * qz

        return np.array(
            [(ww + xx - yy - zz) * x + 2 * ((xy - wz) * y + (xz + wy) * z),
             (ww - xx + yy - zz) * y + 2 * ((xy + wz) * x + (yz - wx) * z),
             (ww - xx - yy + zz) * z + 2 * ((xz - wy) * x + (yz + wx) * y)])

    def rotation_matrix(self):

        qw, qx, qy, qz = self.q[0], self.q[1], self.q[2], self.q[3]

        ww = qw * qw
        xx = qx * qx
        yy = qy * qy
        zz = qz * qz
        wx = qw * qx
        wy = qw * qy
        wz = qw * qz
        xy = qx * qy
        xz = qx * qz
        yz = qy * qz

        return np.array([[ww + xx - yy - zz, 2 * (xy - wz), 2 * (xz + wy)],
                         [2 * (xy + wz), ww - xx + yy - zz, 2 * (yz - wx)],
                         [2 * (xz - wy), 2 * (yz + wx), ww - xx - yy + zz]])

    def axis_angle(self):
        """Returns axis and angle (in radians) for the rotation described
        by this Quaternion"""

        sinth_2 = np.linalg.norm(self.q[1:])

        if sinth_2 == 0:
            # The angle is zero
            theta = 0.0
            n = np.array([0, 0, 1])
        else:
            theta = np.arctan2(sinth_2, self.q[0]) * 2
            n = self.q[1:] / sinth_2

        return n, theta

    def euler_angles(self, mode='zyz'):
        """Return three Euler angles describing the rotation, in radians.
        Mode can be zyz or zxz. Default is zyz."""
        # https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0276302
        if mode == 'zyz':
            a, b, c, d = self.q[0], self.q[3], self.q[2], -self.q[1]
        elif mode == 'zxz':
            a, b, c, d = self.q[0], self.q[3], self.q[1], self.q[2]
        else:
            raise ValueError(f'Invalid Euler angles mode {mode}')

        beta = 2 * np.arccos(
            np.sqrt((a**2 + b**2) / (a**2 + b**2 + c**2 + d**2))
        )
        gap = np.arctan2(b, a)  # (gamma + alpha) / 2
        gam = np.arctan2(d, c)  # (gamma - alpha) / 2
        if np.isclose(beta, 0):
            # gam is meaningless here
            alpha = 0
            gamma = 2 * gap - alpha
        elif np.isclose(beta, np.pi):
            # gap is meaningless here
            alpha = 0
            gamma = 2 * gam + alpha
        else:
            alpha = gap - gam
            gamma = gap + gam

        return np.array([alpha, beta, gamma])

    def arc_distance(self, other):
        """Gives a metric of the distance between two quaternions,
        expressed as 1-|q1.q2|"""

        return 1.0 - np.abs(np.dot(self.q, other.q))

    @staticmethod
    def rotate_byq(q, vector):
        """Apply the rotation matrix to a vector."""
        qw, qx, qy, qz = q[0], q[1], q[2], q[3]
        x, y, z = vector[0], vector[1], vector[2]

        ww = qw * qw
        xx = qx * qx
        yy = qy * qy
        zz = qz * qz
        wx = qw * qx
        wy = qw * qy
        wz = qw * qz
        xy = qx * qy
        xz = qx * qz
        yz = qy * qz

        return np.array(
            [(ww + xx - yy - zz) * x + 2 * ((xy - wz) * y + (xz + wy) * z),
             (ww - xx + yy - zz) * y + 2 * ((xy + wz) * x + (yz - wx) * z),
             (ww - xx - yy + zz) * z + 2 * ((xz - wy) * x + (yz + wx) * y)])

    @staticmethod
    def from_matrix(matrix):
        """Build quaternion from rotation matrix."""
        m = np.array(matrix)
        assert m.shape == (3, 3)

        # Now we need to find out the whole quaternion
        # This method takes into account the possibility of qw being nearly
        # zero, so it picks the stablest solution

        if m[2, 2] < 0:
            if (m[0, 0] > m[1, 1]):
                # Use x-form
                qx = np.sqrt(1 + m[0, 0] - m[1, 1] - m[2, 2]) / 2.0
                fac = 1.0 / (4 * qx)
                qw = (m[2, 1] - m[1, 2]) * fac
                qy = (m[0, 1] + m[1, 0]) * fac
                qz = (m[0, 2] + m[2, 0]) * fac
            else:
                # Use y-form
                qy = np.sqrt(1 - m[0, 0] + m[1, 1] - m[2, 2]) / 2.0
                fac = 1.0 / (4 * qy)
                qw = (m[0, 2] - m[2, 0]) * fac
                qx = (m[0, 1] + m[1, 0]) * fac
                qz = (m[1, 2] + m[2, 1]) * fac
        else:
            if (m[0, 0] < -m[1, 1]):
                # Use z-form
                qz = np.sqrt(1 - m[0, 0] - m[1, 1] + m[2, 2]) / 2.0
                fac = 1.0 / (4 * qz)
                qw = (m[1, 0] - m[0, 1]) * fac
                qx = (m[2, 0] + m[0, 2]) * fac
                qy = (m[1, 2] + m[2, 1]) * fac
            else:
                # Use w-form
                qw = np.sqrt(1 + m[0, 0] + m[1, 1] + m[2, 2]) / 2.0
                fac = 1.0 / (4 * qw)
                qx = (m[2, 1] - m[1, 2]) * fac
                qy = (m[0, 2] - m[2, 0]) * fac
                qz = (m[1, 0] - m[0, 1]) * fac

        return Quaternion(np.array([qw, qx, qy, qz]))

    @staticmethod
    def from_axis_angle(n, theta):
        """Build quaternion from axis (n, vector of 3 components) and angle
        (theta, in radianses)."""

        n = np.array(n, float) / np.linalg.norm(n)
        return Quaternion(np.concatenate([[np.cos(theta / 2.0)],
                                          np.sin(theta / 2.0) * n]))

    @staticmethod
    def from_euler_angles(a, b, c, mode='zyz'):
        """Build quaternion from Euler angles, given in radians. Default
        mode is ZYZ, but it can be set to ZXZ as well."""

        q_a = Quaternion.from_axis_angle([0, 0, 1], a)
        q_c = Quaternion.from_axis_angle([0, 0, 1], c)

        if mode == 'zyz':
            q_b = Quaternion.from_axis_angle([0, 1, 0], b)
        elif mode == 'zxz':
            q_b = Quaternion.from_axis_angle([1, 0, 0], b)
        else:
            raise ValueError(f'Invalid Euler angles mode {mode}')

        return q_c * q_b * q_a



import pytest

from ase.io.formats import ioformats


def test_manually():
    traj = ioformats['traj']
    print(traj)

    outcar = ioformats['vasp-out']
    print(outcar)
    assert outcar.match_name('OUTCAR')
    assert outcar.match_name('something.with.OUTCAR.stuff')


@pytest.fixture()
def excitingtools():
    """If we cannot import excitingtools we skip tests with this fixture."""
    return pytest.importorskip('excitingtools')


@pytest.mark.parametrize('name', ioformats)
def test_ioformat(name, excitingtools):
    """Test getting the full description of each ioformat."""
    ioformat = ioformats[name]
    print(name)
    print('=' * len(name))
    print(ioformat.full_description())
    print()



import numpy as np
import pytest

from ase.build import molecule
from ase.md.velocitydistribution import Stationary, ZeroRotation

norm = np.linalg.norm


@pytest.fixture()
def atoms():
    rng = np.random.RandomState(0)
    atoms = molecule('CH3CH2OH')
    momenta = -0.5 + rng.random((len(atoms), 3))
    atoms.set_momenta(momenta)
    return atoms


@pytest.fixture()
def stationary_atoms(atoms):
    atoms = atoms.copy()
    Stationary(atoms)
    return atoms


def propagate(atoms):
    dt = 0.1
    atoms = atoms.copy()
    velocities = atoms.get_velocities()
    displacement = velocities * dt
    atoms.positions += displacement
    return atoms


def test_stationary(atoms, stationary_atoms):
    assert norm(atoms.get_momenta().sum(axis=0)) > 0.1
    assert norm(stationary_atoms.get_momenta().sum(axis=0)) < 1e-13


def test_stationary_propagate(atoms, stationary_atoms):
    # Test that center of mass is stationary by time propagation
    prop_atoms = propagate(atoms)
    stationary_prop_atoms = propagate(stationary_atoms)
    com = atoms.get_center_of_mass()

    # Center of mass have moved (a bit, at least):
    assert norm(prop_atoms.get_center_of_mass() - com) > 1e-4
    assert norm(stationary_prop_atoms.get_center_of_mass() - com) < 1e-13


def test_zero_rotation(atoms):
    mom1 = atoms.get_angular_momentum()
    ZeroRotation(atoms)
    mom2 = atoms.get_angular_momentum()
    assert norm(mom1) > 0.1
    assert norm(mom2) < 1e-13


# Can we write a "propagation" test for zero rotation as well?



"""Test ase.io.ulm file stuff."""
import numpy as np
import pytest

import ase.io.ulm as ulm


class A:
    def write(self, writer):
        writer.write(x=np.ones((2, 3)))

    @staticmethod
    def read(reader):
        a = A()
        a.x = reader.x
        return a


@pytest.fixture()
def ulmfile(tmp_path):
    path = tmp_path / 'a.ulm'

    with ulm.open(path, 'w') as w:
        w.write(a=A(), y=9)
        w.write(s='abc')
        w.sync()
        w.write(s='abc2')
        w.sync()
        w.write(s='abc3', z=np.ones(7, int))

    return path


class MyFile:
    def __fspath__(self):
        return 'hello'


def test_open_anypathlike():
    # File does not exist, but we still want to test that it is correctly
    # converted into a (nonexistent) path
    with pytest.raises(FileNotFoundError):
        ulm.open(MyFile())


def test_ulm(ulmfile):
    with ulm.open(ulmfile) as r:
        assert r.y == 9
        assert r.s == 'abc'
        assert (A.read(r.a).x == np.ones((2, 3))).all()
        assert (r.a.x == np.ones((2, 3))).all()
        assert r[1].s == 'abc2'
        assert r[2].s == 'abc3'
        assert (r[2].z == np.ones(7)).all()


def test_append(ulmfile):
    path = ulmfile.with_name('b.ulm')
    path.write_bytes(ulmfile.read_bytes())
    with ulm.open(path, 'a') as w:
        assert w.nitems == 3
        w.write(d={'h': [1, 'asdf']})
        w.add_array('psi', (4, 2))
        w.fill(np.ones((1, 2)))
        w.fill(np.ones((1, 2)) * 2)
        w.fill(np.ones((2, 2)) * 3)

    with ulm.open(path, 'r', 3) as r:
        assert r.d['h'] == [1, 'asdf']
    with ulm.open(path) as r:
        assert (r[2].z == np.ones(7)).all()
    with ulm.open(path, index=3) as r:
        psi = r.proxy('psi')[0:3]
        assert (psi == [[1, 1], [2, 2], [3, 3]]).all()


def test_ulm_copy(ulmfile):
    path = ulmfile.with_name('c.ulm')
    ulm.copy(ulmfile, path, exclude={'.a'})

    with ulm.open(path) as r:
        assert 'a' not in r
        assert 'y' in r



import numpy as np

from ase.cell import Cell


def test_niggli_op():

    rng = np.random.RandomState(3)

    for _ in range(5):
        cell = Cell(rng.random((3, 3)))
        print(cell.cellpar())
        rcell, op = cell.niggli_reduce()
        print(op)
        rcell1 = Cell(op.T @ cell)

        rcellpar = rcell.cellpar()
        rcellpar1 = rcell1.cellpar()
        err = np.abs(rcellpar - rcellpar1).max()
        print(err)
        assert err < 1e-10, err



"""
The OPLS io module uses the coordinatetransform.py module under
calculators/lammps/.  This test simply ensures that it uses that module
correctly.  First, it ensures that the interface there hasn't changed.  Second,
it ensures that coordinate system changes are accounted for correctly by
starting with a structure that does not meet LAMMPS' convention for cell
orientation and checking that it gets rotated into the expected configuration
"""
import numpy as np
import pytest

from ase.io.opls import OPLSff, OPLSStructure


@pytest.fixture()
def opls_structure_file_name(datadir):
    return datadir / "opls_structure_ext.xyz"


@pytest.fixture()
def opls_force_field_file_name(datadir):
    """Need to define OPLS parameters for each species in order to be able to
    write a lammps input file"""
    # TODO: This parameter file is being read from the 'testdata' directory,
    #       but it actually already exists under doc/ase/io.  Any sensible way
    #       to prevent duplicating information?
    return str(datadir / "172_defs.par")


def test_opls_write_lammps(opls_structure_file_name,
                           opls_force_field_file_name):

    LAMMPS_FILES_PREFIX = "lmp"

    # Get structure
    atoms = OPLSStructure(opls_structure_file_name)

    # Set up force field object
    with open(opls_force_field_file_name) as fd:
        opls_force_field = OPLSff(fd)

    # Write input files for lammps to current directory
    opls_force_field.write_lammps(atoms, prefix=LAMMPS_FILES_PREFIX)

    # Read the lammps data file
    with open(LAMMPS_FILES_PREFIX + "_atoms") as fd:
        lammps_data = fd.readlines()

    # Locate Atoms block and extract the data for the three atoms in the
    # input structure
    for ind, line in enumerate(lammps_data):
        if line.startswith("Atoms"):
            atom1_data = lammps_data[ind + 2]
            atom2_data = lammps_data[ind + 3]
            atom3_data = lammps_data[ind + 4]
            break

    # Grab positions from data
    pos_indices = slice(4, 7)
    atom1_pos = np.array(atom1_data.split()[pos_indices], dtype=float)
    atom2_pos = np.array(atom2_data.split()[pos_indices], dtype=float)
    atom3_pos = np.array(atom3_data.split()[pos_indices], dtype=float)

    # Check that positions match expected values
    assert atom1_pos == pytest.approx(
        np.array([1.6139, -0.7621, 0.0]), abs=1e-4)
    assert atom2_pos == pytest.approx(np.array([-0.3279, 0.5227, 0]), abs=1e-4)
    assert atom3_pos == pytest.approx(
        np.array([-0.96, 0.5809, 0.88750]), abs=1e-4)



from ase.geometry.dimensionality.interval_analysis import (
    analyze_kintervals as analyze_dimensionality,
)
from ase.geometry.dimensionality.isolation import isolate_components

__all__ = ['analyze_dimensionality', 'isolate_components']



import numpy as np
import pytest

from ase import Atoms
from ase.build import bulk
from ase.io import read

from .test_mustem import make_STO_atoms


def test_write_read_cycle_xyz_prismatic():
    """Check writing and reading a xtl mustem file."""
    # Reproduce the SI100.XYZ file distributed with prismatic
    atoms = bulk('Si', cubic=True)
    atoms.set_array('occupancies', np.ones(len(atoms)))
    rng = np.random.RandomState(42)
    atoms.set_array('debye_waller_factors', 0.62 + 0.1 * rng.random(len(atoms)))

    filename = 'SI100.XYZ'
    atoms.write(filename=filename, format='prismatic',
                comments='one unit cell of 100 silicon')

    atoms_loaded = read(filename=filename, format='prismatic')

    np.testing.assert_allclose(atoms.positions, atoms_loaded.positions)
    np.testing.assert_allclose(atoms.cell, atoms_loaded.cell)
    np.testing.assert_allclose(atoms.get_array('occupancies'),
                               atoms_loaded.get_array('occupancies'))
    np.testing.assert_allclose(atoms.get_array('debye_waller_factors'),
                               atoms_loaded.get_array('debye_waller_factors'),
                               rtol=1E-5)


def test_write_error():
    """Check missing parameter when writing xyz prismatic file."""
    atoms_Si100 = bulk('Si', cubic=True)
    atoms_STO = make_STO_atoms()
    filename = 'SI100.XYZ'

    with pytest.raises(ValueError):
        # DW not provided
        atoms_Si100.write(filename, format='prismatic')

    # Write file with DW provided as scalar
    atoms_Si100.write(filename, format='prismatic',
                      debye_waller_factors=0.076)

    # Write file with DW provided as dict
    atoms_Si100.write(filename, format='prismatic',
                      debye_waller_factors={'Si': 0.076})

    STO_DW_dict = {'Sr': 0.62, 'O': 0.73, 'Ti': 0.43}
    STO_DW_dict_Ti_missing = {key: STO_DW_dict[key] for key in ['Sr', 'O']}

    with pytest.raises(ValueError):
        # DW missing keys
        atoms_STO.write(filename, format='prismatic',
                        debye_waller_factors=STO_DW_dict_Ti_missing)

    atoms_STO.write(filename, format='prismatic',
                    debye_waller_factors=STO_DW_dict)

    with pytest.raises(ValueError):
        # Raise an error if the unit cell is not defined.
        atoms4 = Atoms(['Sr', 'Ti', 'O', 'O', 'O'],
                       positions=[[0, 0, 0],
                                  [0.5, 0.5, 0.5],
                                  [0.5, 0.5, 0],
                                  [0.5, 0, 0.5],
                                  [0, 0.5, 0.5]])
        atoms4.write(filename, format='prismatic',
                     debye_waller_factors=STO_DW_dict)



import warnings
from dataclasses import dataclass

import numpy as np

spin_error = (
    'The spin keyword is no longer supported.  Please call the function '
    'with the energies corresponding to the desired spins.')
_deprecated = object()


def get_band_gap(calc, direct=False, spin=_deprecated):
    warnings.warn('Please use ase.dft.bandgap.bandgap() instead!')
    gap, (s1, k1, _n1), (s2, k2, _n2) = bandgap(calc, direct, spin=spin)
    ns = calc.get_number_of_spins()
    if ns == 2:
        return gap, (s1, k1), (s2, k2)
    return gap, k1, k2


@dataclass
class GapInfo:
    eigenvalues: np.ndarray

    def __post_init__(self):
        self._gapinfo = _bandgap(self.eigenvalues, direct=False)
        self._direct_gapinfo = _bandgap(self.eigenvalues, direct=True)

    @classmethod
    def fromcalc(cls, calc):
        kpts = calc.get_ibz_k_points()
        nk = len(kpts)
        ns = calc.get_number_of_spins()
        eigenvalues = np.array([[calc.get_eigenvalues(kpt=k, spin=s)
                                 for k in range(nk)]
                                for s in range(ns)])

        efermi = calc.get_fermi_level()
        return cls(eigenvalues - efermi)

    def gap(self):
        return self._gapinfo

    def direct_gap(self):
        return self._direct_gapinfo

    @property
    def is_metallic(self) -> bool:
        return self._gapinfo[0] == 0.0

    @property
    def gap_is_direct(self) -> bool:
        """Whether the direct and indirect gaps are the same transition."""
        return self._gapinfo[1:] == self._direct_gapinfo[1:]

    def description(self, *, ibz_kpoints=None) -> str:
        """Return human-friendly description of direct/indirect gap.

        If ibz_k_points are given, coordinates are printed as well."""
        from typing import List

        lines: List[str] = []
        add = lines.append

        def skn(skn):
            """Convert k-point indices (s, k, n) to string."""
            description = 's={}, k={}, n={}'.format(*skn)
            if ibz_kpoints is not None:
                coordtxt = '[{:.2f}, {:.2f}, {:.2f}]'.format(
                    *ibz_kpoints[skn[1]])
                description = f'{description}, {coordtxt}'
            return f'({description})'

        gap, skn1, skn2 = self.gap()
        direct_gap, skn_direct1, skn_direct2 = self.direct_gap()

        if self.is_metallic:
            add('No gap')
        else:
            add(f'Gap: {gap:.3f} eV')
            add('Transition (v -> c):')
            add(f'  {skn(skn1)} -> {skn(skn2)}')

        if self.gap_is_direct:
            add('No difference between direct/indirect transitions')
        else:
            add('Direct/indirect transitions are different')
            add(f'Direct gap: {direct_gap:.3f} eV')
            if skn_direct1[0] == skn_direct2[0]:
                add(f'Transition at: {skn(skn_direct1)}')
            else:
                transition = skn((f'{skn_direct1[0]}->{skn_direct2[0]}',
                                  *skn_direct1[1:]))
                add(f'Transition at: {transition}')

        return '\n'.join(lines)


def bandgap(calc=None, direct=False, spin=_deprecated,
            eigenvalues=None, efermi=None, output=None, kpts=None):
    """Calculates the band-gap.

    Parameters:

    calc: Calculator object
        Electronic structure calculator object.
    direct: bool
        Calculate direct band-gap.
    eigenvalues: ndarray of shape (nspin, nkpt, nband) or (nkpt, nband)
        Eigenvalues.
    efermi: float
        Fermi level (defaults to 0.0).

    Returns a (gap, p1, p2) tuple where p1 and p2 are tuples of indices of the
    valence and conduction points (s, k, n).

    Example:

    >>> gap, p1, p2 = bandgap(silicon.calc)
    >>> print(gap, p1, p2)
    1.2 (0, 0, 3), (0, 5, 4)
    >>> gap, p1, p2 = bandgap(silicon.calc, direct=True)
    >>> print(gap, p1, p2)
    3.4 (0, 0, 3), (0, 0, 4)
    """

    if spin is not _deprecated:
        raise RuntimeError(spin_error)

    if calc:
        kpts = calc.get_ibz_k_points()
        nk = len(kpts)
        ns = calc.get_number_of_spins()
        eigenvalues = np.array([[calc.get_eigenvalues(kpt=k, spin=s)
                                 for k in range(nk)]
                                for s in range(ns)])
        if efermi is None:
            efermi = calc.get_fermi_level()

    efermi = efermi or 0.0

    gapinfo = GapInfo(eigenvalues - efermi)

    e_skn = gapinfo.eigenvalues
    if eigenvalues.ndim == 2:
        e_skn = e_skn[np.newaxis]  # spinors

    if not np.isfinite(e_skn).all():
        raise ValueError('Bad eigenvalues!')

    gap, (s1, k1, n1), (s2, k2, n2) = _bandgap(e_skn, direct)

    if eigenvalues.ndim != 3:
        p1 = (k1, n1)
        p2 = (k2, n2)
    else:
        p1 = (s1, k1, n1)
        p2 = (s2, k2, n2)

    return gap, p1, p2


def _bandgap(e_skn, direct):
    """Helper function."""
    ns, nk, nb = e_skn.shape
    s1 = s2 = k1 = k2 = n1 = n2 = None

    N_sk = (e_skn < 0.0).sum(2)  # number of occupied bands

    # Check for bands crossing the fermi-level
    if ns == 1:
        if np.ptp(N_sk[0]) > 0:
            return 0.0, (None, None, None), (None, None, None)
    else:
        if (np.ptp(N_sk, axis=1) > 0).any():
            return 0.0, (None, None, None), (None, None, None)

    if (N_sk == 0).any() or (N_sk == nb).any():
        raise ValueError('Too few bands!')

    e_skn = np.array([[e_skn[s, k, N_sk[s, k] - 1:N_sk[s, k] + 1]
                       for k in range(nk)]
                      for s in range(ns)])
    ev_sk = e_skn[:, :, 0]  # valence band
    ec_sk = e_skn[:, :, 1]  # conduction band

    if ns == 1:
        s1 = 0
        s2 = 0
        gap, k1, k2 = find_gap(ev_sk[0], ec_sk[0], direct)
        n1 = N_sk[0, 0] - 1
        n2 = n1 + 1
        return gap, (0, k1, n1), (0, k2, n2)

    gap, k1, k2 = find_gap(ev_sk.ravel(), ec_sk.ravel(), direct)
    if direct:
        # Check also spin flips:
        for s in [0, 1]:
            g, k, _ = find_gap(ev_sk[s], ec_sk[1 - s], direct)
            if g < gap:
                gap = g
                k1 = k + nk * s
                k2 = k + nk * (1 - s)

    if gap > 0.0:
        s1, k1 = divmod(k1, nk)
        s2, k2 = divmod(k2, nk)
        n1 = N_sk[s1, k1] - 1
        n2 = N_sk[s2, k2]
        return gap, (s1, k1, n1), (s2, k2, n2)
    return 0.0, (None, None, None), (None, None, None)


def find_gap(ev_k, ec_k, direct):
    """Helper function."""
    if direct:
        gap_k = ec_k - ev_k
        k = gap_k.argmin()
        return gap_k[k], k, k
    kv = ev_k.argmax()
    kc = ec_k.argmin()
    return ec_k[kc] - ev_k[kv], kv, kc



"""Test module for explicitly unittesting parts of the VASP calculator"""

import os
import sys

import numpy as np
import pytest

from ase import Atoms
from ase.build import molecule
from ase.calculators.calculator import (
    CalculatorSetupError,
    get_calculator_class,
)
from ase.calculators.vasp import Vasp
from ase.calculators.vasp.vasp import (
    check_atoms,
    check_atoms_type,
    check_cell,
    check_pbc,
)


@pytest.fixture(name="atoms")
def fixture_atoms():
    return molecule('H2', vacuum=5, pbc=True)


@pytest.fixture(autouse=True)
def always_mock_calculate(mock_vasp_calculate):
    """No tests in this module may execute VASP"""
    return


def test_verify_no_run():
    """Verify that we get an error if we try and execute the calculator,
    due to the fixture.
    """
    calc = Vasp()
    with pytest.raises(AssertionError):
        calc._run()


def test_check_atoms(atoms):
    """Test checking atoms passes for a good atoms object"""
    check_atoms(atoms)
    check_pbc(atoms)
    check_cell(atoms)


@pytest.mark.parametrize(
    'bad_atoms',
    [
        None,
        'a_string',
        # We cannot handle lists of atoms either
        [molecule('H2', vacuum=5)],
    ])
def test_not_atoms(bad_atoms):
    """Check that passing in objects which are not
    actually Atoms objects raises a setup error """

    with pytest.raises(CalculatorSetupError):
        check_atoms_type(bad_atoms)
    with pytest.raises(CalculatorSetupError):
        check_atoms(bad_atoms)


@pytest.mark.parametrize('pbc', [
    3 * [False],
    [True, False, True],
    [False, True, False],
])
def test_bad_pbc(atoms, pbc):
    """Test handling of PBC"""
    atoms.pbc = pbc

    check_cell(atoms)  # We have a cell, so this should not raise

    # Check that our helper functions raises the expected error
    with pytest.raises(CalculatorSetupError):
        check_pbc(atoms)
    with pytest.raises(CalculatorSetupError):
        check_atoms(atoms)

    # Check we also raise in the calculator when launching
    # a calculation, but before VASP is actually executed
    calc = Vasp()
    atoms.calc = calc
    with pytest.raises(CalculatorSetupError):
        atoms.get_potential_energy()


def test_vasp_no_cell(testdir):
    """Check missing cell handling."""
    # Molecules come with no unit cell
    atoms = molecule('CH4')
    # We should not have a cell
    assert atoms.cell.rank == 0

    with pytest.raises(CalculatorSetupError):
        check_cell(atoms)
    with pytest.raises(CalculatorSetupError):
        check_atoms(atoms)

    with pytest.raises(RuntimeError):
        atoms.write('POSCAR')

    calc = Vasp()
    atoms.calc = calc
    with pytest.raises(CalculatorSetupError):
        atoms.get_total_energy()


def test_spinpol_vs_ispin():
    """Test if `spinpol` is consistent with `ispin`"""
    atoms = molecule("O2")
    atoms.set_initial_magnetic_moments([1.0, 1.0])

    calc = Vasp(ispin=1)
    calc._set_spinpol(atoms)
    assert not calc.spinpol

    calc = Vasp(ispin=2)
    calc._set_spinpol(atoms)
    assert calc.spinpol

    # when `ispin` is not specified, `spinpol` is determined by `magmom`
    calc = Vasp()
    calc._set_spinpol(atoms)
    assert calc.spinpol


class TestReadMagneticMoments:
    """Test if the "magnetization (x)" block in OUTCAR is parsed correctly."""

    # from `bulk("Fe", "bcc", a=2.8630354989499160, cubic=True)`
    lines_spd = [
        " magnetization (x)\n",
        "\n",
        "# of ion       s       p       d       tot\n",
        "------------------------------------------\n",
        "    1       -0.009  -0.045   2.369   2.315\n",
        "    2       -0.009  -0.045   2.369   2.315\n",
        "--------------------------------------------------\n"
        "tot         -0.019  -0.089   4.738   4.630\n",
    ]

    # from `bulk("Ga", "bcc", a=4.0692361730014719, cubic=True)`
    lines_spdf = [
        " magnetization (x)\n",
        "\n",
        "# of ion       s       p       d       f       tot\n",
        "--------------------------------------------------\n",
        "    1        0.013   0.007   0.329   6.877   7.227\n",
        "    2        0.013   0.007   0.329   6.877   7.227\n",
        "--------------------------------------------------\n",
        "tot          0.026   0.014   0.659  13.755  14.454\n",
    ]

    @pytest.mark.parametrize(
        'lines, magmoms_ref', (
            (lines_spd, (2.315, 2.315)),
            (lines_spdf, (7.227, 7.227)),
        )
    )
    def test(self, lines, magmoms_ref):
        """Test"""
        calc = Vasp()
        # dummy atoms
        calc.atoms = Atoms(
            ("Fe", "Gd"),
            positions=((0.0, 0.0, 0.0), (0.5, 0.5, 0.5)),
        )
        calc.resort = [0, 1]
        magmoms = calc._read_magnetic_moments(lines)
        np.testing.assert_allclose(magmoms, magmoms_ref)


def test_read_magnetic_moment():
    """Test if the magnetization line in OUTCAR is parsed correctly."""
    # ISPIN 1
    ln1 = " number of electron       8.0000000 magnetization \n"
    assert Vasp()._read_magnetic_moment([ln1]) == 0.0
    # ISPIN 2
    ln2 = " number of electron       8.0000000 magnetization       2.0000000\n"
    assert Vasp()._read_magnetic_moment([ln2]) == 2.0


def test_vasp_name():
    """Test the calculator class has the expected name"""
    expected = 'vasp'
    assert Vasp.name == expected  # Test class attribute
    assert Vasp().name == expected  # Ensure instance attribute hasn't changed


def test_vasp_get_calculator():
    cls_ = get_calculator_class('vasp')

    assert cls_ == Vasp

    # Test we get the correct calculator when loading from name
    assert get_calculator_class(Vasp.name) == cls_


@pytest.mark.parametrize('envvar', Vasp.env_commands)
def test_make_command_envvar(envvar, monkeypatch, clear_vasp_envvar):
    """Test making a command based on the environment variables"""
    # Environment should be cleared by the "clear_vasp_envvar" fixture
    assert envvar not in os.environ
    cmd_str = 'my command'
    monkeypatch.setenv(envvar, cmd_str)
    calc = Vasp()

    cmd = calc.make_command()
    if envvar == 'VASP_SCRIPT':
        # This envvar uses the `sys.exe` to create the command
        exe = sys.executable
        assert cmd == f'{exe} {cmd_str}'
    else:
        # We just use the exact string
        assert cmd == cmd_str


def test_make_command_no_envvar(monkeypatch, clear_vasp_envvar):
    """Test we raise when making a command with not enough information"""
    # Environment should be cleared by the "clear_vasp_envvar" fixture
    calc = Vasp()
    with pytest.raises(CalculatorSetupError):
        calc.make_command()


def test_make_command_explicit(monkeypatch):
    """Test explicitly passing a command to the calculator"""
    # The command should be whatever we put in, if we explicitly set something
    # envvars should not matter
    for envvar in Vasp.env_commands:
        # Populate the envvars with some strings, they should not matter
        monkeypatch.setenv(envvar, 'something')
    calc = Vasp()
    my_cmd = 'my command'
    cmd = calc.make_command(my_cmd)
    assert cmd == my_cmd


def test_as_dict():
    calc = Vasp(lreal=False, xc="pbe")
    dct = calc.asdict()
    inputs = dct["inputs"]
    assert inputs["lreal"] is False
    assert inputs["xc"] == "pbe"



from collections.abc import Iterable


def write_incar(directory, parameters, header=None):
    incar_string = generate_incar_lines(parameters)
    with open(f"{directory}/INCAR", "w") as incar:
        if header is not None:
            incar.write(header + "\n")
        incar.write(incar_string)


def generate_incar_lines(parameters):
    if isinstance(parameters, str):
        return parameters
    elif parameters is None:
        return ""
    else:
        incar_lines = []
        for item in parameters.items():
            incar_lines += list(generate_line(*item))
        # Adding a newline at the end of the file
        return "\n".join(incar_lines) + "\n"


def generate_line(key, value, num_spaces=0):
    indent = " " * num_spaces
    if isinstance(value, str):
        if value.find("\n") != -1:
            value = '"' + value + '"'
        yield indent + f"{key.upper()} = {value}"
    elif isinstance(value, dict):
        yield indent + f"{key.upper()} {{"
        for item in value.items():
            yield from generate_line(*item, num_spaces + 4)
        yield indent + "}"
    elif isinstance(value, Iterable):
        yield indent + f"{key.upper()} = {' '.join(str(x) for x in value)}"
    else:
        yield indent + f"{key.upper()} = {value}"



from ase.build import bulk
from ase.dft.kpoints import bandpath


def test_makebandpath():
    atoms = bulk('Au')
    cell = atoms.cell

    path0 = bandpath('GXL', cell)
    print(path0)
    path1 = bandpath([[0., 0., 0.], [.5, .5, .5]], cell, npoints=50)
    print(path1)
    path2 = bandpath([[0., 0., 0.], [.5, .5, .5], [.1, .2, .3]], cell,
                     npoints=50,
                     special_points={'G': [0., 0., 0.]})
    print(path2)



from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from ase.ga import set_raw_score
from ase.ga.standard_comparators import (
    EnergyComparator,
    InteratomicDistanceComparator,
    RawScoreComparator,
    SequentialComparator,
)


def test_standardcomparator():

    a1 = Atoms('AgAgAg', positions=[[0, 0, 0], [1.5, 0, 0], [1.5, 1.5, 0]])
    a2 = Atoms('AgAgAg', positions=[[0, 0, 0], [1.4, 0, 0], [1.5, 1.5, 0]])

    e1 = 1.0
    e2 = 0.8

    a1.calc = SinglePointCalculator(a1, energy=e1)
    a2.calc = SinglePointCalculator(a2, energy=e2)

    comp1 = InteratomicDistanceComparator(n_top=3,
                                          pair_cor_cum_diff=0.03,
                                          pair_cor_max=0.7,
                                          dE=0.3)
    assert comp1.looks_like(a1, a2)

    comp2 = InteratomicDistanceComparator(n_top=3,
                                          pair_cor_cum_diff=0.03,
                                          pair_cor_max=0.7,
                                          dE=0.15)
    assert not comp2.looks_like(a1, a2)

    comp3 = InteratomicDistanceComparator(n_top=3,
                                          pair_cor_cum_diff=0.02,
                                          pair_cor_max=0.7,
                                          dE=0.3)
    assert not comp3.looks_like(a1, a2)

    hard_E_comp = EnergyComparator(dE=1.0)
    assert hard_E_comp.looks_like(a1, a2)

    soft_E_comp = EnergyComparator(dE=.01)
    assert not soft_E_comp.looks_like(a1, a2)

    set_raw_score(a1, .1)
    set_raw_score(a2, .27)

    rs_comp = RawScoreComparator(0.15)
    assert not rs_comp.looks_like(a1, a2)

    comp1 = SequentialComparator([hard_E_comp, rs_comp], [0, 0])
    assert not comp1.looks_like(a1, a2)

    comp2 = SequentialComparator([hard_E_comp, rs_comp], [0, 1])
    assert comp2.looks_like(a1, a2)



import pytest
from numpy.testing import assert_allclose

from ase.build import molecule
from ase.calculators.mopac import MOPAC
from ase.io.trajectory import Trajectory
from ase.optimize import BFGS


@pytest.mark.calculator_lite()
@pytest.mark.calculator()
def test_mopac(mopac_factory):
    """Test H2 molecule atomization with MOPAC."""
    # Unrestricted Hartree-Fock; enable magmom calc
    h2 = molecule('H2',
                  calculator=mopac_factory.calc(label='h2',
                                                task='1SCF GRADIENTS UHF'))
    with Trajectory('h2.traj', mode='w') as traj:
        BFGS(h2, trajectory=traj).run(fmax=0.01)
    e2 = h2.get_potential_energy()
    h1 = h2.copy()
    del h1[1]
    h1.set_initial_magnetic_moments([1])
    h1.calc = mopac_factory.calc(label='h1')
    e1 = h1.get_potential_energy()
    d = h2.get_distance(0, 1)
    ea = 2 * e1 - e2
    print(d, ea)
    assert abs(d - 0.759) < 0.001
    assert abs(ea - 5.907) < 0.001
    h2o = molecule('H2O', calculator=mopac_factory.calc(label='h2o',
                                                        tasks='GRADIENTS'))
    h2o.get_potential_energy()
    print('dipole:', h2o.get_dipole_moment())
    atoms = MOPAC.read_atoms('h2', profile=h2o.calc.profile)
    print('magmom:', atoms.calc.get_magnetic_moment())
    print('PM7 homo lumo:', atoms.calc.get_homo_lumo_levels())
    atoms.calc.set(method='AM1')
    atoms.get_potential_energy()
    print('AM1 homo lumo:', atoms.calc.get_homo_lumo_levels())
    calc = mopac_factory.calc(restart='h1')
    print('magmom:', calc.get_magnetic_moment())


@pytest.mark.calculator_lite()
def test_mopac_forces_consistent(mopac_factory):
    """Check MOPAC forces follow Newton's 3rd Law"""

    ch4 = molecule('CH4')
    ch4.rattle()
    ch4.calc = mopac_factory.calc(task='1SCF GRADIENTS', method='PM7')
    forces = ch4.get_forces()

    assert_allclose(forces.sum(axis=0), [0, 0, 0], atol=1e-7)



"""
IO support for the Gaussian cube format.

See the format specifications on:
http://local.wasp.uwa.edu.au/~pbourke/dataformats/cube/
"""

import time

import numpy as np

from ase.atoms import Atoms
from ase.io import read
from ase.units import Bohr

ATOMS = 'atoms'
CASTEP = 'castep'
DATA = 'data'


def write_cube(file_obj, atoms, data=None, origin=None, comment=None):
    """Function to write a cube file.

    file_obj: str or file object
        File to which output is written.
    atoms: Atoms
        The Atoms object specifying the atomic configuration.
    data : 3-dim numpy array, optional (default = None)
        Array containing volumetric data as e.g. electronic density
    origin : 3-tuple
        Origin of the volumetric data (units: Angstrom)
    comment : str, optional (default = None)
        Comment for the first line of the cube file.
    """

    if data is None:
        data = np.ones((2, 2, 2))
    data = np.asarray(data)

    if data.dtype == complex:
        data = np.abs(data)

    if comment is None:
        comment = "Cube file from ASE, written on " + time.strftime("%c")
    else:
        comment = comment.strip()
    file_obj.write(comment)

    file_obj.write("\nOUTER LOOP: X, MIDDLE LOOP: Y, INNER LOOP: Z\n")

    if origin is None:
        origin = np.zeros(3)
    else:
        origin = np.asarray(origin) / Bohr

    file_obj.write(
        "{:5}{:12.6f}{:12.6f}{:12.6f}\n".format(
            len(atoms), *origin))

    for i in range(3):
        n = data.shape[i]
        d = atoms.cell[i] / n / Bohr
        file_obj.write("{:5}{:12.6f}{:12.6f}{:12.6f}\n".format(n, *d))

    positions = atoms.positions / Bohr
    numbers = atoms.numbers
    for Z, (x, y, z) in zip(numbers, positions):
        file_obj.write(
            "{:5}{:12.6f}{:12.6f}{:12.6f}{:12.6f}\n".format(
                Z, 0.0, x, y, z)
        )

    data.tofile(file_obj, sep="\n", format="%e")


def read_cube(file_obj, read_data=True, program=None, verbose=False):
    """Read atoms and data from CUBE file.

    file_obj : str or file
        Location to the cube file.
    read_data : boolean
        If set true, the actual cube file content, i.e. an array
        containing the electronic density (or something else )on a grid
        and the dimensions of the corresponding voxels are read.
    program: str
        Use program='castep' to follow the PBC convention that first and
        last voxel along a direction are mirror images, thus the last
        voxel is to be removed.  If program=None, the routine will try
        to catch castep files from the comment lines.
    verbose : bool
        Print some more information to stdout.

    Returns a dict with the following keys:

    * 'atoms': Atoms object
    * 'data' : (Nx, Ny, Nz) ndarray
    * 'origin': (3,) ndarray, specifying the cube_data origin.
    * 'spacing': (3, 3) ndarray, representing voxel size
    """

    readline = file_obj.readline
    line = readline()  # the first comment line
    line = readline()  # the second comment line

    # The second comment line *CAN* contain information on the axes
    # But this is by far not the case for all programs
    axes = []
    if "OUTER LOOP" in line.upper():
        axes = ["XYZ".index(s[0]) for s in line.upper().split()[2::3]]
    if not axes:
        axes = [0, 1, 2]

    # castep2cube files have a specific comment in the second line ...
    if "castep2cube" in line:
        program = CASTEP
        if verbose:
            print("read_cube identified program: castep")

    # Third line contains actual system information:
    line = readline().split()
    num_atoms = int(line[0])

    # num_atoms can be negative.
    # Negative num_atoms indicates we have extra data to parse after
    # the coordinate information.
    has_labels = num_atoms < 0
    num_atoms = abs(num_atoms)

    # There is an optional last field on this line which indicates
    # the number of values at each point. It is typically 1 (the default)
    # in which case it can be omitted, but it may also be > 1,
    # for example if there are multiple orbitals stored in the same cube.
    num_val = int(line[4]) if len(line) == 5 else 1

    # Origin around which the volumetric data is centered
    # (at least in FHI aims):
    origin = np.array([float(x) * Bohr for x in line[1:4:]])

    cell = np.empty((3, 3))
    shape = []
    spacing = np.empty((3, 3))

    # The upcoming three lines contain the cell information
    for i in range(3):
        n, x, y, z = (float(s) for s in readline().split())
        shape.append(int(n))

        # different PBC treatment in castep, basically the last voxel row is
        # identical to the first one
        if program == CASTEP:
            n -= 1
        cell[i] = n * Bohr * np.array([x, y, z])
        spacing[i] = np.array([x, y, z]) * Bohr
    pbc = [(v != 0).any() for v in cell]

    numbers = np.empty(num_atoms, int)
    positions = np.empty((num_atoms, 3))
    for i in range(num_atoms):
        line = readline().split()
        numbers[i] = int(line[0])
        positions[i] = [float(s) for s in line[2:]]

    positions *= Bohr

    atoms = Atoms(numbers=numbers, positions=positions, cell=cell, pbc=pbc)

    # CASTEP will always have PBC, although the cube format does not
    # contain this kind of information
    if program == CASTEP:
        atoms.pbc = True

    dct = {ATOMS: atoms}
    labels = []

    # If we originally had a negative num_atoms, parse the extra fields now.
    # The first field of the first line tells us how many other fields there
    # are to parse, but we have to guess how many rows this information is
    # split over.
    if has_labels:
        # Can't think of a more elegant way of doing this...
        fields = readline().split()
        nfields = int(fields[0])
        labels.extend(fields[1:])

        while len(labels) < nfields:
            fields = readline().split()
            labels.extend(fields)

    labels = [int(x) for x in labels]

    if read_data:
        # Cube files can contain more than one density,
        # so we need to be a little bit careful about where one ends
        # and the next begins.
        raw_volume = [float(s) for s in file_obj.read().split()]
        # Split each value at each point into a separate list.
        raw_volumes = [np.array(raw_volume[offset::num_val])
                       for offset in range(num_val)]

        datas = []

        # Adjust each volume in turn.
        for data in raw_volumes:
            data = data.reshape(shape)
            if axes != [0, 1, 2]:
                data = data.transpose(axes).copy()

            if program == CASTEP:
                # Due to the PBC applied in castep2cube, the last entry
                # along each dimension equals the very first one.
                data = data[:-1, :-1, :-1]

            datas.append(data)

        datas = np.array(datas)

        dct[DATA] = datas[0]
        dct["origin"] = origin
        dct["spacing"] = spacing
        dct["labels"] = labels
        dct["datas"] = datas

    return dct


def read_cube_data(filename):
    """Wrapper function to read not only the atoms information from a cube file
    but also the contained volumetric data.
    """
    dct = read(filename, format="cube", read_data=True, full_output=True)
    return dct[DATA], dct[ATOMS]



from ase import Atoms
from ase.constraints import FixBondLengths


def test_fixbonds():
    """Test Atoms.__delitem__ with FixAtoms constraint."""

    a = Atoms('H3')
    a.constraints = FixBondLengths([(1, 2)])
    assert (a[:].constraints[0].pairs == [(1, 2)]).all()
    assert (a[1:].constraints[0].pairs == [(0, 1)]).all()
    assert len(a[2:].constraints) == 0
    assert len(a[1:2].constraints) == 0
    assert len(a[:2].constraints) == 0
    assert len(a[:1].constraints) == 0

    # Execise Atoms.__init__:
    Atoms(a)



# flake8: noqa
"""Definition of the XrDebye class.

This module defines the XrDebye class for calculation
of X-ray scattering properties from atomic cluster
using Debye formula.
Also contains routine for calculation of atomic form factors and
X-ray wavelength dict.
"""

from math import acos, cos, exp, pi, sin, sqrt

import numpy as np

from ase.data import atomic_numbers

# Table (1) of
# D. WAASMAIER AND A. KIRFEL, Acta Cryst. (1995). A51, 416-431
waasmaier = {
    #       a1         b1         a2         b2         a3         b3          a4         b4         a5         b5         c
    'C': [2.657506, 14.780758, 1.078079, 0.776775, 1.490909, 42.086843, -4.241070, -0.000294, 0.713791, 0.239535, 4.297983],
    'N': [11.893780, 0.000158, 3.277479, 10.232723, 1.858092, 30.344690, 0.858927, 0.656065, 0.912985, 0.217287, -11.804902],
    'O': [2.960427, 14.182259, 2.5088111, 5.936858, 0.637053, 0.112726, 0.722838, 34.958481, 1.142756, 0.390240, 0.027014],
    'P': [1.950541, 0.908139, 4.146930, 27.044953, 1.494560, 0.071280, 1.522042, 67.520190, 5.729711, 1.981173, 0.155233],
    'S': [6.372157, 1.514347, 5.154568, 22.092528, 1.473732, 0.061373, 1.635073, 55.445176, 1.209372, 0.646925, 0.154722],
    'Cl': [1.446071, 0.052357, 6.870609, 1.193165, 6.151801, 18.343416, 1.750347, 46.398394, 0.634168, 0.401005, 0.146773],
    'Ni': [13.521865, 4.077277, 6.947285, 0.286763, 3.866028, 14.622634, 2.135900, 71.966078, 4.284731, 0.004437, -2.762697],
    'Cu': [14.014192, 3.738280, 4.784577, 0.003744, 5.056806, 13.034982, 1.457971, 72.554793, 6.932996, 0.265666, -3.774477],
    'Pd': [6.121511, 0.062549, 4.784063, 0.784031, 16.631683, 8.751391, 4.318258, 34.489983, 13.246773, 0.784031, 0.883099],
    'Ag': [6.073874, 0.055333, 17.155437, 7.896512, 4.173344, 28.443739, 0.852238, 110.376108, 17.988685, 0.716809, 0.756603],
    'Pt': [31.273891, 1.316992, 18.445441, 8.797154, 17.063745, 0.124741, 5.555933, 40.177994, 1.575270, 1.316997, 4.050394],
    'Au': [16.777389, 0.122737, 19.317156, 8.621570, 32.979682, 1.256902, 5.595453, 38.008821, 10.576854, 0.000601, -6.279078],
}

wavelengths = {
    'CuKa1': 1.5405981,
    'CuKa2': 1.54443,
    'CuKb1': 1.39225,
    'WLa1': 1.47642,
    'WLa2': 1.48748
}


class XrDebye:
    """
    Class for calculation of XRD or SAXS patterns.
    """

    def __init__(self, atoms, wavelength, damping=0.04,
                 method='Iwasa', alpha=1.01, warn=True):
        """
        Initilize the calculation of X-ray diffraction patterns

        Parameters:

        atoms: ase.Atoms
            atoms object for which calculation will be performed.

        wavelength: float, Angstrom
            X-ray wavelength in Angstrom. Used for XRD and to setup dumpings.

        damping : float, Angstrom**2
            thermal damping factor parameter (B-factor).

        method: {'Iwasa'}
            method of calculation (damping and atomic factors affected).

            If set to 'Iwasa' than angular damping and q-dependence of
            atomic factors are used.

            For any other string there will be only thermal damping
            and constant atomic factors (`f_a(q) = Z_a`).

        alpha: float
            parameter for angular damping of scattering intensity.
            Close to 1.0 for unplorized beam.

        warn: boolean
            flag to show warning if atomic factor can't be calculated
        """
        self.wavelength = wavelength
        self.damping = damping
        self.mode = ''
        self.method = method
        self.alpha = alpha
        self.warn = warn

        self.twotheta_list = []
        self.q_list = []
        self.intensity_list = []

        self.atoms = atoms
        # TODO: setup atomic form factors if method != 'Iwasa'

    def set_damping(self, damping):
        """ set B-factor for thermal damping """
        self.damping = damping

    def get(self, s):
        r"""Get the powder x-ray (XRD) scattering intensity
        using the Debye-Formula at single point.

        Parameters:

        s: float, in inverse Angstrom
            scattering vector value (`s = q / 2\pi`).

        Returns:
            Intensity at given scattering vector `s`.
        """

        pre = exp(-self.damping * s**2 / 2)

        if self.method == 'Iwasa':
            sinth = self.wavelength * s / 2.
            positive = 1. - sinth**2
            if positive < 0:
                positive = 0
            costh = sqrt(positive)
            cos2th = cos(2. * acos(costh))
            pre *= costh / (1. + self.alpha * cos2th**2)

        f = {}

        def atomic(symbol):
            """
            get atomic factor, using cache.
            """
            if symbol not in f:
                if self.method == 'Iwasa':
                    f[symbol] = self.get_waasmaier(symbol, s)
                else:
                    f[symbol] = atomic_numbers[symbol]
            return f[symbol]

        I = 0.
        fa = []  # atomic factors list
        for a in self.atoms:
            fa.append(atomic(a.symbol))

        pos = self.atoms.get_positions()  # positions of atoms
        fa = np.array(fa)  # atomic factors array

        for i in range(len(self.atoms)):
            vr = pos - pos[i]
            I += np.sum(fa[i] * fa * np.sinc(2 * s *
                        np.sqrt(np.sum(vr * vr, axis=1))))

        return pre * I

    def get_waasmaier(self, symbol, s):
        r"""Scattering factor for free atoms.

        Parameters:

        symbol: string
            atom element symbol.

        s: float, in inverse Angstrom
            scattering vector value (`s = q / 2\pi`).

        Returns:
            Intensity at given scattering vector `s`.

        Note:
            for hydrogen will be returned zero value."""
        if symbol == 'H':
            # XXXX implement analytical H
            return 0
        elif symbol in waasmaier:
            abc = waasmaier[symbol]
            f = abc[10]
            s2 = s * s
            for i in range(5):
                f += abc[2 * i] * exp(-abc[2 * i + 1] * s2)
            return f
        if self.warn:
            print('<xrdebye::get_atomic> Element', symbol, 'not available')
        return 0

    def calc_pattern(self, x=None, mode='XRD', verbose=False):
        r"""
        Calculate X-ray diffraction pattern or
        small angle X-ray scattering pattern.

        Parameters:

        x: float array
            points where intensity will be calculated.
            XRD - 2theta values, in degrees;
            SAXS - q values in 1/A
            (`q = 2 \pi \cdot s = 4 \pi \sin( \theta) / \lambda`).
            If ``x`` is ``None`` then default values will be used.

        mode: {'XRD', 'SAXS'}
            the mode of calculation: X-ray diffraction (XRD) or
            small-angle scattering (SAXS).

        Returns:
            list of intensities calculated for values given in ``x``.
        """
        self.mode = mode.upper()
        assert (mode in ['XRD', 'SAXS'])

        result = []
        if mode == 'XRD':
            if x is None:
                self.twotheta_list = np.linspace(15, 55, 100)
            else:
                self.twotheta_list = x
            self.q_list = []
            if verbose:
                print('#2theta\tIntensity')
            for twotheta in self.twotheta_list:
                s = 2 * sin(twotheta * pi / 180 / 2.0) / self.wavelength
                result.append(self.get(s))
                if verbose:
                    print(f'{twotheta:.3f}\t{result[-1]:f}')
        elif mode == 'SAXS':
            if x is None:
                self.twotheta_list = np.logspace(-3, -0.3, 100)
            else:
                self.q_list = x
            self.twotheta_list = []
            if verbose:
                print('#q\tIntensity')
            for q in self.q_list:
                s = q / (2 * pi)
                result.append(self.get(s))
                if verbose:
                    print(f'{q:.4f}\t{result[-1]:f}')
        self.intensity_list = np.array(result)
        return self.intensity_list

    def write_pattern(self, filename):
        """ Save calculated data to file specified by ``filename`` string."""
        with open(filename, 'w') as fd:
            self._write_pattern(fd)

    def _write_pattern(self, fd):
        fd.write('# Wavelength = %f\n' % self.wavelength)
        if self.mode == 'XRD':
            x, y = self.twotheta_list, self.intensity_list
            fd.write('# 2theta \t Intesity\n')
        elif self.mode == 'SAXS':
            x, y = self.q_list, self.intensity_list
            fd.write('# q(1/A)\tIntesity\n')
        else:
            raise Exception('No data available, call calc_pattern() first.')

        for i in range(len(x)):
            fd.write(f'  {x[i]:f}\t{y[i]:f}\n')

    def plot_pattern(self, filename=None, show=False, ax=None):
        """ Plot XRD or SAXS depending on filled data

        Uses Matplotlib to plot pattern. Use *show=True* to
        show the figure and *filename='abc.png'* or
        *filename='abc.eps'* to save the figure to a file.

        Returns:
            ``matplotlib.axes.Axes`` object."""

        import matplotlib.pyplot as plt

        if ax is None:
            plt.clf()  # clear figure
            ax = plt.gca()

        if self.mode == 'XRD':
            x, y = np.array(self.twotheta_list), np.array(self.intensity_list)
            ax.plot(x, y / np.max(y), '.-')
            ax.set_xlabel('2$\\theta$')
            ax.set_ylabel('Intensity')
        elif self.mode == 'SAXS':
            x, y = np.array(self.q_list), np.array(self.intensity_list)
            ax.loglog(x, y / np.max(y), '.-')
            ax.set_xlabel('q, 1/Angstr.')
            ax.set_ylabel('Intensity')
        else:
            raise Exception('No data available, call calc_pattern() first')

        if show:
            plt.show()
        if filename is not None:
            fig = ax.get_figure()
            fig.savefig(filename)

        return ax



"""Tests for NeighborList"""
import numpy as np
import pytest

from ase import Atoms
from ase.build import bulk
from ase.neighborlist import (
    NeighborList,
    NewPrimitiveNeighborList,
    PrimitiveNeighborList,
)


@pytest.mark.parametrize(
    'primitive',
    [PrimitiveNeighborList, NewPrimitiveNeighborList],
)
@pytest.mark.parametrize('bothways', [False, True])
@pytest.mark.parametrize('self_interaction', [False, True])
@pytest.mark.parametrize('sorted', [False, True])
def test_unique(
    sorted: bool,
    self_interaction: bool,
    bothways: bool,
    primitive,
):
    """Test if there are no duplicates in the neighbor lists"""
    atoms = Atoms('H2', positions=[(0, 0, 0), (0, 0, 1)])
    nl = NeighborList(
        [0.5, 0.5],
        skin=0.1,
        sorted=sorted,
        self_interaction=self_interaction,
        bothways=bothways,
        primitive=primitive,
    )
    nl.update(atoms)
    tmp = []
    for i in range(len(atoms)):
        neighbors, offsets = nl.get_neighbors(i)
        tmp += [(i, n, *o) for n, o in zip(neighbors, offsets)]
    assert len(set(tmp)) == len(tmp)


def count(nl: NeighborList, atoms: Atoms):
    """Count the numbers of neighboring atoms for all the atoms

    Returns
    -------
    d : float
        Sum of distances over all nearest-neighbor pairs
    c : npt.NDArray[np.int_]
        Numbers of neighboring atoms for all the atoms
    """
    c = np.zeros(len(atoms), int)
    R = atoms.get_positions()
    cell = atoms.get_cell()
    d = 0.0
    for a in range(len(atoms)):
        i, offsets = nl.get_neighbors(a)
        for j in i:
            c[j] += 1
        c[a] += len(i)
        d += (((R[i] + np.dot(offsets, cell) - R[a])**2).sum(1)**0.5).sum()
    return d, c


# scipy sparse uses matrix subclass internally
@pytest.mark.filterwarnings('ignore:the matrix subclass')
@pytest.mark.slow()
@pytest.mark.parametrize('sorted', [False, True])
def test_supercell(sorted):
    """Test if NeighborList works for a supercell as expected"""
    atoms = Atoms(numbers=range(10),
                  cell=[(0.2, 1.2, 1.4),
                        (1.4, 0.1, 1.6),
                        (1.3, 2.0, -0.1)])
    atoms.set_scaled_positions(3 * np.random.random((10, 3)) - 1)

    for p1 in range(2):
        for p2 in range(2):
            for p3 in range(2):
                # print(p1, p2, p3)
                atoms.set_pbc((p1, p2, p3))
                cutoffs = atoms.numbers * 0.2 + 0.5
                nl = NeighborList(cutoffs, skin=0.0, sorted=sorted)
                nl.update(atoms)
                d, c = count(nl, atoms)

                atoms2 = atoms.repeat((p1 + 1, p2 + 1, p3 + 1))
                cutoffs2 = atoms2.numbers * 0.2 + 0.5
                nl2 = NeighborList(cutoffs2, skin=0.0, sorted=sorted)
                nl2.update(atoms2)
                d2, c2 = count(nl2, atoms2)

                c2.shape = (-1, 10)  # row: images, column: atoms

                # if the sum of nearest-neighbor distances gets larger
                # according to the supercell size
                dd = d * (p1 + 1) * (p2 + 1) * (p3 + 1) - d2
                assert abs(dd) < 1e-10

                # if each repeated image has the same numbers of neighbors
                assert not (c2 - c).any()


def test_H2():
    h2 = Atoms('H2', positions=[(0, 0, 0), (0, 0, 1)])
    nl = NeighborList([0.5, 0.5], skin=0.1, sorted=True, self_interaction=False)
    nl2 = NeighborList([0.5, 0.5], skin=0.1, sorted=True,
                       self_interaction=False,
                       primitive=NewPrimitiveNeighborList)
    assert nl2.update(h2)
    assert nl.update(h2)
    assert not nl.update(h2)
    assert (nl.get_neighbors(0)[0] == [1]).all()
    m = np.zeros((2, 2))
    m[0, 1] = 1
    assert np.array_equal(nl.get_connectivity_matrix(sparse=False), m)
    assert np.array_equal(nl.get_connectivity_matrix(sparse=True).todense(), m)
    assert np.array_equal(nl.get_connectivity_matrix().todense(),
                          nl2.get_connectivity_matrix().todense())

    h2[1].z += 0.09
    assert not nl.update(h2)
    assert (nl.get_neighbors(0)[0] == [1]).all()

    h2[1].z += 0.09
    assert nl.update(h2)
    assert (nl.get_neighbors(0)[0] == []).all()
    assert nl.nupdates == 2


def test_H2_shape_and_type():
    h2 = Atoms('H2', positions=[(0, 0, 0), (0, 0, 1)])
    nl = NeighborList([0.1, 0.1], skin=0.1, bothways=True,
                      self_interaction=False)
    assert nl.update(h2)
    assert nl.get_neighbors(0)[1].shape == (0, 3)
    assert nl.get_neighbors(0)[1].dtype == int


def test_fcc():
    x = bulk('X', 'fcc', a=2**0.5)

    nl = NeighborList([0.5], skin=0.01, bothways=True, self_interaction=False)
    nl.update(x)
    assert len(nl.get_neighbors(0)[0]) == 12

    nl = NeighborList([0.5] * 27, skin=0.01, bothways=True,
                      self_interaction=False)
    nl.update(x * (3, 3, 3))
    for a in range(27):
        assert len(nl.get_neighbors(a)[0]) == 12
    assert not np.any(nl.get_neighbors(13)[1])


def test_use_scaled_positions():
    c = 0.0058
    for NeighborListClass in [PrimitiveNeighborList, NewPrimitiveNeighborList]:
        nl = NeighborListClass([c, c],
                               skin=0.0,
                               sorted=True,
                               self_interaction=False,
                               use_scaled_positions=True)
        nl.update([True, True, True],
                  np.eye(3) * 7.56,
                  np.array([[0, 0, 0],
                            [0, 0, 0.99875]]))
        n0, d0 = nl.get_neighbors(0)
        n1, d1 = nl.get_neighbors(1)
        # != is xor
        assert (np.all(n0 == [0]) and np.all(d0 == [0, 0, 1])) != \
            (np.all(n1 == [1]) and np.all(d1 == [0, 0, -1]))


def test_empty_neighbor_list():
    # Test empty neighbor list
    nl = PrimitiveNeighborList([])
    nl.update([True, True, True],
              np.eye(3) * 7.56,
              np.zeros((0, 3)))


@pytest.mark.parametrize('bothways', [False, True])
@pytest.mark.parametrize('self_interaction', [False, True])
@pytest.mark.parametrize('sort', [False, True])
def test_equivalence_of_primitive_classes(sort, self_interaction, bothways):
    """Test if two primitive neighbor-list classes make the same naighbors"""
    # diamond structure in the primitive cell
    pbc_c = np.array([True, True, True])
    cell_cv = np.array([[0., 3.37316113, 3.37316113],
                        [3.37316113, 0., 3.37316113],
                        [3.37316113, 3.37316113, 0.]])
    spos_ac = np.array([[0., 0., 0.],
                        [0.25, 0.25, 0.25]])
    natoms = len(spos_ac)

    cutoff_a = np.array([8.0, 8.0])

    info = [[] for _ in range(2)]  # neighbor info collector for each primitive
    primitives = [PrimitiveNeighborList, NewPrimitiveNeighborList]
    for ip, primitive in enumerate(primitives):
        nl = primitive(
            cutoff_a,
            skin=0.0,
            sorted=sort,
            self_interaction=self_interaction,
            bothways=bothways,
            use_scaled_positions=True,
        )
        nl.update(pbc_c, cell_cv, spos_ac)

        # collect neighbor info into a list of tuples
        # each tuple has the form (i1, i2, o1, o2, o3)
        # i1: 1st atom
        # i2: 2nd atom
        # o1: offset along 1st cell vector
        # o2: offset along 2nd cell vector
        # o3: offset along 3rd cell vector
        for i in range(natoms):
            info[ip].extend([(i, n, *o) for n, o in zip(*nl.get_neighbors(i))])

    def reverse(t: tuple):
        return t[1], t[0], -t[2], -t[3], -t[4]

    for ip in range(2):
        # (i1, i2, +o1, +o2, +o3) and (i2, i1, -o1, -o2, -o3) is the same pair
        # the following guarantees i0 <= i1
        info[ip] = [t if t[0] <= t[1] else reverse(t) for t in info[ip]]
        info[ip] = sorted(info[ip])  # sort by i1, i2, o1, o2, o3

    # check if the both primitive classes provide the same neighbors
    assert np.all(info[0] == info[1])


def test_small_cell_and_large_cutoff():
    # See: https://gitlab.com/ase/ase/-/issues/441
    cutoff = 50

    atoms = bulk('Cu', 'fcc', a=3.6)
    atoms *= (2, 2, 2)
    atoms.set_pbc(False)
    radii = cutoff * np.ones(len(atoms.get_atomic_numbers()))

    neighborhood_new = NeighborList(
        radii, skin=0.0, self_interaction=False, bothways=True,
        primitive=NewPrimitiveNeighborList
    )
    neighborhood_old = NeighborList(
        radii, skin=0.0, self_interaction=False, bothways=True,
        primitive=PrimitiveNeighborList
    )

    neighborhood_new.update(atoms)
    neighborhood_old.update(atoms)

    n0, d0 = neighborhood_new.get_neighbors(0)
    n1, d1 = neighborhood_old.get_neighbors(0)

    assert np.all(n0 == n1)
    assert np.all(d0 == d1)



from ase import Atom, Atoms
from ase.calculators.lj import LennardJones
from ase.constraints import FixBondLength


def test_dimer():
    dimer = Atoms([Atom('X', (0, 0, 0)),
                   Atom('X', (0, 0, 1))],
                  calculator=LennardJones(),
                  constraint=FixBondLength(0, 1))
    print(dimer.get_forces())
    print(dimer.positions)
    dimer.positions[:] += 0.1
    print(dimer.positions)
    dimer.positions[:, 2] += 5.1
    print(dimer.positions)
    dimer.positions[:] = [(1, 2, 3), (4, 5, 6)]
    print(dimer.positions)
    dimer.set_positions([(1, 2, 3), (4, 5, 6.2)])
    print(dimer.positions)



# flake8: noqa
import numpy as np

from ase.transport.calculators import TransportCalculator

# Aux. function to write data to a text file.


def write(fname, xs, ys):
    with open(fname, 'w') as fd:
        for x, y in zip(xs, ys):
            print(x, y, file=fd)


def test_transport_calculator(testdir):
    H_lead = np.zeros([4, 4])

    # On-site energies are zero
    for i in range(4):
        H_lead[i, i] = 0.0

    # Nearest neighbor hopping is -1.0
    for i in range(3):
        H_lead[i, i + 1] = -1.0
        H_lead[i + 1, i] = -1.0

    # Next-nearest neighbor hopping is 0.2
    for i in range(2):
        H_lead[i, i + 2] = 0.2
        H_lead[i + 2, i] = 0.2

    H_scat = np.zeros([6, 6])
    # Principal layers on either side of S
    H_scat[:2, :2] = H_lead[:2, :2]
    H_scat[-2:, -2:] = H_lead[:2, :2]

    # Scattering region
    H_scat[2, 2] = 0.0
    H_scat[3, 3] = 0.0
    H_scat[2, 3] = -0.8
    H_scat[3, 2] = -0.8

    # External coupling
    H_scat[1, 2] = 0.2
    H_scat[2, 1] = 0.2
    H_scat[3, 4] = 0.2
    H_scat[4, 3] = 0.2

    energies = np.arange(-3, 3, 0.02)
    tcalc = TransportCalculator(h=H_scat,
                                h1=H_lead,
                                eta=0.02,
                                energies=energies)

    T = tcalc.get_transmission()
    tcalc.set(pdos=[2, 3])
    pdos = tcalc.get_pdos()

    tcalc.set(dos=True)
    dos = tcalc.get_dos()

    write('T.dat', tcalc.energies, T)
    write('pdos0.dat', tcalc.energies, pdos[0])
    write('pdos1.dat', tcalc.energies, pdos[1])

    # subdiagonalize
    h_rot, s_rot, eps, u = tcalc.subdiagonalize_bfs([2, 3], apply=True)
    T_rot = tcalc.get_transmission()
    dos_rot = tcalc.get_dos()
    pdos_rot = tcalc.get_pdos()

    write('T_rot.dat', tcalc.energies, T_rot)
    write('pdos0_rot.dat', tcalc.energies, pdos_rot[0])
    write('pdos1_rot.dat', tcalc.energies, pdos_rot[1])

    print('Subspace eigenvalues:', eps)
    assert sum(abs(eps - (-0.8, 0.8))) < 2.0e-15, 'Subdiagonalization. error'
    print('Max deviation of T after the rotation:', np.abs(T - T_rot).max())
    assert max(abs(T - T_rot)) < 2.0e-15, 'Subdiagonalization. error'

    # remove coupling
    h_cut, s_cut = tcalc.cutcoupling_bfs([2], apply=True)
    T_cut = tcalc.get_transmission()
    dos_cut = tcalc.get_dos()
    pdos_cut = tcalc.get_pdos()

    write('T_cut.dat', tcalc.energies, T_cut)
    write('pdos0_cut.dat', tcalc.energies, pdos_cut[0])
    write('pdos1_cut.dat', tcalc.energies, pdos_cut[1])



from os.path import isfile, join

from ase.build import molecule
from ase.calculators.onetep import Onetep


def test_onetep(onetep_factory):
    mol = molecule('H2O')
    mol.center(8)
    calc = Onetep(label='water')
    # Tests conducted with the JTH PAW data set.
    # http://www.abinit.org/downloads/PAW2
    prefix = 'hello'
    h_path = join(prefix, 'H.abinit')
    o_path = join(prefix, 'O.abinit')
    if not (isfile(h_path) and isfile(o_path)):
        raise Exception("""You must supply PAW data sets for
            hydrogen and oxygen to run this test.
            Please see http://www.abinit.org/downloads/PAW2
            for suitable data. ONETEP takes PAW data sets in the
            abinit format. I need H.abinit and O.abinit""")
    calc.set_pseudos([('H', h_path), ('O', o_path)])
    calc.set(paw=True, xc='PBE', cutoff_energy='400 eV')
    mol.calc = calc

    energy = mol.get_total_energy()
    ref_energy = -470.852068717
    assert abs(energy - ref_energy) < 1e-6



# flake8: noqa

"""Test the CP2K ASE calulator.

http://www.cp2k.org
Author: Ole Schuett <ole.schuett@mat.ethz.ch>
"""

from io import StringIO

import pytest

from ase.io.cp2k import read_cp2k_restart


@pytest.fixture
def inp():
    return StringIO("""\
 # Version information for this restart file
 &MOTION
   &CELL_OPT
     MAX_ITER  800
   &END CELL_OPT
 &FORCE_EVAL
   &DFT
     UKS  T
   &END DFT
   &SUBSYS
     &CELL
       A     1.4436982360095069E+01    0.0000000000000000E+00    0.0000000000000000E+00
       B    -1.2546391461121697E+01    8.0771799263415858E+00    0.0000000000000000E+00
       PERIODIC  XY
       MULTIPLE_UNIT_CELL  1 1 1
     &END CELL
     &COORD
C   -3.7242617044497828E+00    7.9038234645202037E+00    3.4613477913211641E+00
C1   -2.9068950543864061E+00    4.7668576748644087E+00    5.8444011777519380E+00
cu1    1.6455807102639135E+00    5.4728919446731368E+00    5.7625128629895181E+00
     &END COORD
   &END SUBSYS
   &PRINT
   &END PRINT
 &END FORCE_EVAL
""")


def test_restart(inp):
    mol = read_cp2k_restart(inp)
    assert len(mol) == 3
    assert (mol.get_pbc() == [True, True, False]).all()



import numpy as np
import pytest

from ase.utils.filecache import (
    CombinedJSONCache,
    Locked,
    MultiFileJSONCache,
    MultiFileULMCache,
)

pytestmark = pytest.mark.usefixtures('testdir')


@pytest.fixture(params=['json', 'ulm'])
def cache(request):
    caches = {'json': MultiFileJSONCache('cache_json'),
              'ulm': MultiFileULMCache('cache_ulm')}
    return caches[request.param]


def sample_dict():
    return {'hello': [1, 2, 3], 'world': 'grumble'}


def test_basic(cache):
    assert len(cache) == 0

    cache['hello'] = 'grumble'
    assert len(cache) == 1
    assert 'hello' in cache

    grumble = cache.pop('hello')
    assert grumble == 'grumble'
    assert 'hello' not in cache
    assert len(cache) == 0


def test_numpy_array(cache):
    assert len(cache) == 0
    cache['a'] = np.array([3.4, 2.4, 1.4j])
    cache['b'] = np.array([3.4, 2.4, 1.4])
    assert len(cache) == 2
    assert 'a' in cache
    aa = cache.pop('a')
    bb = cache.pop('b')
    assert np.allclose(aa, np.array([3.4, 2.4, 1.4j]))
    assert np.allclose(bb, np.array([3.4, 2.4, 1.4]))
    assert len(cache) == 0


@pytest.mark.parametrize('dct', [{}, sample_dict()])
def test_cache(dct, cache):
    cache.update(dct)
    assert dict(cache) == dct


def test_combine(cache):
    dct = sample_dict()
    cache.update(dct)
    combined = cache.combine()
    assert dict(combined) == dct


def test_split():
    dct = sample_dict()
    combined = CombinedJSONCache.dump_cache('cache', dct)
    assert dict(combined) == dct
    cache = combined.split()
    assert dict(cache) == dct
    assert len(combined) == 0


def test_lock(cache):
    with cache.lock('hello'):
        # When element is locked but nothing is written, the
        # cache is defined to "contain" None
        assert 'hello' in cache
        assert cache['hello'] is None

        # Other keys should function as normal:
        cache['xx'] = 1
        assert cache['xx'] == 1


def test_already_locked(cache):
    with cache.lock('hello') as handle:
        assert handle is not None
        with cache.lock('hello') as otherhandle:
            assert otherhandle is None

        with pytest.raises(Locked):
            cache['hello'] = 'world'


def test_no_overwrite_combine(cache):
    cache.combine()
    with pytest.raises(RuntimeError, match='Already exists'):
        cache.combine()



from math import cos, pi, sin

from ase import Atoms
from ase.build import add_adsorbate, fcc100, fcc111
from ase.cluster import wulff_construction
from ase.constraints import FixAtoms
from ase.db import connect
from ase.lattice.cubic import FaceCenteredCubic


def get_systems():
    systems = []

    cell = (5, 5, 5)
    atoms = Atoms('H2', [(0, 0, 0), (0, 0, 1.4)], cell=cell)
    atoms.center()
    systems.append((atoms, 'Hydrogen molecule'))

    atoms = FaceCenteredCubic(
        directions=[[1, -1, 0], [1, 1, 0], [0, 0, 1]],
        size=(2, 2, 2),
        symbol='Cu',
        pbc=(1, 1, 1))
    atoms.rattle(stdev=0.1, seed=42)
    systems.append((atoms, 'Shaken bulk copper'))

    a = 2.70
    c = 1.59 * a

    slab = Atoms('2Cu', [(0., 0., 0.), (1 / 3., 1 / 3., -0.5 * c)],
                 tags=(0, 1),
                 pbc=(1, 1, 0))
    slab.set_cell([(a, 0, 0),
                   (a / 2, 3**0.5 * a / 2, 0),
                   (0, 0, 1)])
    slab.center(vacuum=3, axis=2)
    mask = [a.tag == 1 for a in slab]
    slab.set_constraint(FixAtoms(mask=mask))
    systems.append((slab, 'Distorted Cu(111) surface'))

    zpos = cos(134.3 / 2.0 * pi / 180.0) * 1.197
    xpos = sin(134.3 / 2.0 * pi / 180.0) * 1.19
    co = Atoms('CO', positions=[(-xpos + 1.2, 0, -zpos),
                                (-xpos + 1.2, -1.1, -zpos)])
    slab = fcc111('Au', size=(2, 2, 2), orthogonal=True)
    add_adsorbate(slab, co, 1.5, 'bridge')
    slab.center(vacuum=6, axis=2)
    slab.set_pbc((True, True, False))
    constraint = FixAtoms(mask=[a.tag == 2 for a in slab])
    slab.set_constraint(constraint)
    systems.append((slab, 'CO on Au(111) surface'))

    atoms = Atoms(symbols='C5H12',
                  cell=[16.83752497, 12.18645905, 11.83462179],
                  positions=[[5.90380523, 5.65545388, 5.91569796],
                             [7.15617518, 6.52907738, 5.91569796],
                             [8.41815022, 5.66384716, 5.92196554],
                             [9.68108996, 6.52891016, 5.91022362],
                             [10.93006206, 5.65545388, 5.91569796],
                             [5.00000011, 6.30002353, 5.9163716],
                             [5.88571848, 5.0122839, 6.82246859],
                             [5.88625613, 5.01308931, 5.01214155],
                             [7.14329342, 7.18115393, 6.81640316],
                             [7.14551332, 7.17200869, 5.00879027],
                             [8.41609966, 5.00661165, 5.02355167],
                             [8.41971183, 5.0251482, 6.83462168],
                             [9.69568096, 7.18645894, 6.8078633],
                             [9.68914668, 7.16663649, 5.00000011],
                             [10.95518898, 5.02163182, 6.8289018],
                             [11.83752486, 6.29836826, 5.90274952],
                             [10.94464142, 5.00000011, 5.01802495]])
    systems.append((atoms, 'Pentane molecule'))

    slab = fcc100('Cu', size=(2, 2, 2), vacuum=3.5)
    add_adsorbate(slab, 'C', 1.5, 'hollow')
    mask = [a.tag > 1 for a in slab]
    constraint = FixAtoms(mask=mask)
    slab.set_constraint(constraint)
    systems.append((slab, 'C/Cu(100)'))

    surfaces = [(1, 0, 0), (1, 1, 0), (1, 1, 1)]
    esurf = [0.9151, 0.9771, 0.7953]  # Surface energies
    size = 10  # number of atoms
    atoms = wulff_construction('Al', surfaces, esurf, size, 'fcc',
                               rounding='above')
    atoms.center(vacuum=6)
    atoms.rattle(0.2)
    systems.append((atoms, 'Alumninum cluster'))
    return systems


def create_database():
    systems = get_systems()
    db = connect('systems.db', append=False)
    for atoms, description in systems:
        name = atoms.get_chemical_formula()
        db.write(atoms, description=description, name=name)


if __name__ == '__main__':
    create_database()



from ase.atoms import Atoms
from ase.calculators.crystal import CRYSTAL
from ase.optimize import BFGS


def test_molecule(crystal_factory):
    with open('basis', 'w') as fd:
        fd.write("""6 4
    0 0 6 2.0 1.0
     3048.0 0.001826
     456.4 0.01406
     103.7 0.06876
     29.23 0.2304
     9.349 0.4685
     3.189 0.3628
    0 1 2 4.0 1.0
     3.665 -0.3959 0.2365
     0.7705 1.216 0.8606
    0 1 1 0.0 1.0
     0.26 1.0 1.0
    0 3 1 0.0 1.0
     0.8 1.0
    """)

    geom = Atoms('OHH',
                 positions=[(0, 0, 0), (1, 0, 0), (0, 1, 0)])

    geom.calc = CRYSTAL(
        label='water',
        guess=True,
        basis='sto-3g',
        xc='PBE',
        otherkeys=['scfdir', 'anderson',
                   ['maxcycles', '500'],
                   ['toldee', '6'],
                   ['tolinteg', '7 7 7 7 14'],
                   ['fmixing', '90']])

    with BFGS(geom) as opt:
        opt.run(fmax=0.05)

    final_energy = geom.get_potential_energy()
    assert abs(final_energy + 2047.34531091) < 1.0



from math import sqrt

import pytest

from ase import Atom, Atoms
from ase.calculators.emt import EMT
from ase.constraints import FixAtoms
from ase.io import read
from ase.optimize import QuasiNewton


@pytest.mark.optimize()
def test_replay(testdir):
    # Distance between Cu atoms on a (100) surface:
    d = 3.6 / sqrt(2)
    a = Atoms('Cu',
              positions=[(0, 0, 0)],
              cell=(d, d, 1.0),
              pbc=(True, True, False))
    a *= (2, 2, 1)  # 2x2 (100) surface-cell

    # Approximate height of Ag atom on Cu(100) surfece:
    h0 = 2.0
    a += Atom('Ag', (d / 2, d / 2, h0))

    constraint = FixAtoms(range(len(a) - 1))
    a.calc = EMT()
    a.set_constraint(constraint)

    with QuasiNewton(a, trajectory='AgCu1.traj', logfile='AgCu1.log') as dyn1:
        dyn1.run(fmax=0.1)

    a = read('AgCu1.traj')
    a.calc = EMT()
    print(a.constraints)

    with QuasiNewton(a, trajectory='AgCu2.traj', logfile='AgCu2.log') as dyn2:
        dyn2.replay_trajectory('AgCu1.traj')
        dyn2.run(fmax=0.01)



"""TAB-completion sub-command and update helper funtion.

Run this when ever options are changed::

    python3 -m ase.cli.completion
"""

import sys
from pathlib import Path
from typing import Dict, List, Tuple

# Path of the complete.py script:
path = Path(__file__).with_name('complete.py')


class CLICommand:
    """Add tab-completion for Bash.

    Will show the command that needs to be added to your '~/.bashrc file.
    """
    cmd = f'complete -o default -C "{sys.executable} {path}" ase'

    @staticmethod
    def add_arguments(parser):
        pass

    @staticmethod
    def run(args):
        cmd = CLICommand.cmd
        print(cmd)


def update(path: Path,
           subcommands: List[Tuple[str, str]],
           test: bool = False) -> None:
    """Update commands dict in complete.py.

    Use test=True to test that no changes are needed.

    Refactor with care!  This function is also used by GPAW.
    """

    import textwrap
    from importlib import import_module

    dct: Dict[str, List[str]] = {}

    class Subparser:
        def __init__(self, command):
            self.command = command
            dct[command] = []

        def add_argument(self, *args, **kwargs):
            dct[command].extend(arg for arg in args
                                if arg.startswith('-'))

        def add_mutually_exclusive_group(self, required=False):
            return self

    for command, module_name in subcommands:
        module = import_module(module_name)
        module.CLICommand.add_arguments(Subparser(command))

    txt = 'commands = {'
    for command, opts in sorted(dct.items()):
        txt += "\n    '" + command + "':\n        ["
        if opts:
            txt += '\n'.join(textwrap.wrap("'" + "', '".join(opts) + "'],",
                                           width=65,
                                           break_on_hyphens=False,
                                           subsequent_indent='         '))
        else:
            txt += '],'
    txt = txt[:-1] + '}\n'

    with path.open() as fd:
        lines = fd.readlines()

    a = lines.index('# Beginning of computer generated data:\n')
    b = lines.index('# End of computer generated data\n')

    if test:
        if ''.join(lines[a + 1:b]) != txt:
            raise ValueError(
                'Please update ase/cli/complete.py using '
                '"python3 -m ase.cli.completion".')
    else:
        lines[a + 1:b] = [txt]
        new = path.with_name('complete.py.new')
        with new.open('w') as fd:
            print(''.join(lines), end='', file=fd)
        new.rename(path)
        path.chmod(0o775)


if __name__ == '__main__':
    from ase.cli.main import commands
    update(path, commands)



import numpy as np

import ase  # Annotations
from ase.calculators.calculator import PropertyNotImplementedError
from ase.utils import jsonable


def calculate_band_structure(atoms, path=None, scf_kwargs=None,
                             bs_kwargs=None, kpts_tol=1e-6, cell_tol=1e-6):
    """Calculate band structure.

    The purpose of this function is to abstract a band structure calculation
    so the workflow does not depend on the calculator.

    First trigger SCF calculation if necessary, then set arguments
    on the calculator for band structure calculation, then return
    calculated band structure.

    The difference from get_band_structure() is that the latter
    expects the calculation to already have been done."""
    if path is None:
        path = atoms.cell.bandpath()

    from ase.lattice import celldiff  # Should this be a method on cell?
    if any(path.cell.any(1) != atoms.pbc):
        raise ValueError('The band path\'s cell, {}, does not match the '
                         'periodicity {} of the atoms'
                         .format(path.cell, atoms.pbc))
    cell_err = celldiff(path.cell, atoms.cell.uncomplete(atoms.pbc))
    if cell_err > cell_tol:
        raise ValueError('Atoms and band path have different unit cells.  '
                         'Please reduce atoms to standard form.  '
                         'Cell lengths and angles are {} vs {}'
                         .format(atoms.cell.cellpar(), path.cell.cellpar()))

    calc = atoms.calc
    if calc is None:
        raise ValueError('Atoms have no calculator')

    if scf_kwargs is not None:
        calc.set(**scf_kwargs)

    # Proposed standard mechanism for calculators to advertise that they
    # use the bandpath keyword to handle band structures rather than
    # a double (SCF + BS) run.
    use_bandpath_kw = getattr(calc, 'accepts_bandpath_keyword', False)
    if use_bandpath_kw:
        calc.set(bandpath=path)
        atoms.get_potential_energy()
        return calc.band_structure()

    atoms.get_potential_energy()

    if hasattr(calc, 'get_fermi_level'):
        # What is the protocol for a calculator to tell whether
        # it has fermi_energy?
        eref = calc.get_fermi_level()
    else:
        eref = 0.0

    if bs_kwargs is None:
        bs_kwargs = {}

    calc.set(kpts=path, **bs_kwargs)
    calc.results.clear()  # XXX get rid of me

    # Calculators are too inconsistent here:
    # * atoms.get_potential_energy() will fail when total energy is
    #   not in results after BS calculation (Espresso)
    # * calc.calculate(atoms) doesn't ask for any quantity, so some
    #   calculators may not calculate anything at all
    # * 'bandstructure' is not a recognized property we can ask for
    try:
        atoms.get_potential_energy()
    except PropertyNotImplementedError:
        pass

    ibzkpts = calc.get_ibz_k_points()
    kpts_err = np.abs(path.kpts - ibzkpts).max()
    if kpts_err > kpts_tol:
        raise RuntimeError('Kpoints of calculator differ from those '
                           'of the band path we just used; '
                           'err={} > tol={}'.format(kpts_err, kpts_tol))

    bs = get_band_structure(atoms, path=path, reference=eref)
    return bs


def get_band_structure(atoms=None, calc=None, path=None, reference=None):
    """Create band structure object from Atoms or calculator."""
    # path and reference are used internally at the moment, but
    # the exact implementation will probably change.  WIP.
    #
    # XXX We throw away info about the bandpath when we create the calculator.
    # If we have kept the bandpath, we can provide it as an argument here.
    # It would be wise to check that the bandpath kpoints are the same as
    # those stored in the calculator.
    atoms = atoms if atoms is not None else calc.atoms
    calc = calc if calc is not None else atoms.calc

    kpts = calc.get_ibz_k_points()

    energies = []
    for s in range(calc.get_number_of_spins()):
        energies.append([calc.get_eigenvalues(kpt=k, spin=s)
                         for k in range(len(kpts))])
    energies = np.array(energies)

    if path is None:
        from ase.dft.kpoints import (
            BandPath,
            find_bandpath_kinks,
            resolve_custom_points,
        )
        standard_path = atoms.cell.bandpath(npoints=0)
        # Kpoints are already evaluated, we just need to put them into
        # the path (whether they fit our idea of what the path is, or not).
        #
        # Depending on how the path was established, the kpoints might
        # be valid high-symmetry points, but since there are multiple
        # high-symmetry points of each type, they may not coincide
        # with ours if the bandpath was generated by another code.
        #
        # Here we hack it so the BandPath has proper points even if they
        # come from some weird source.
        #
        # This operation (manually hacking the bandpath) is liable to break.
        # TODO: Make it available as a proper (documented) bandpath method.
        kinks = find_bandpath_kinks(atoms.cell, kpts, eps=1e-5)
        pathspec, special_points = resolve_custom_points(
            kpts[kinks], standard_path.special_points, eps=1e-5)
        path = BandPath(standard_path.cell,
                        kpts=kpts,
                        path=pathspec,
                        special_points=special_points)

    # XXX If we *did* get the path, now would be a good time to check
    # that it matches the cell!  Although the path can only be passed
    # because we internally want to not re-evaluate the Bravais
    # lattice type.  (We actually need an eps parameter, too.)

    if reference is None:
        # Fermi level should come from the GS calculation, not the BS one!
        reference = calc.get_fermi_level()

    if reference is None:
        # Fermi level may not be available, e.g., with non-Fermi smearing.
        # XXX Actually get_fermi_level() should raise an error when Fermi
        # level wasn't available, so we should fix that.
        reference = 0.0

    return BandStructure(path=path,
                         energies=energies,
                         reference=reference)


class BandStructurePlot:
    def __init__(self, bs):
        self.bs = bs
        self.ax = None
        self.xcoords = None

    def plot(self, ax=None, emin=-10, emax=5, filename=None,
             show=False, ylabel=None, colors=None, point_colors=None,
             label=None, loc=None,
             cmap=None, cmin=-1.0, cmax=1.0, sortcolors=False,
             colorbar=True, clabel='$s_z$', cax=None,
             **plotkwargs):
        """Plot band-structure.

        ax: Axes
            MatPlotLib Axes object.  Will be created if not supplied.
        emin, emax: float
            Minimum and maximum energy above reference.
        filename: str
            If given, write image to a file.
        show: bool
            Show the image (not needed in notebooks).
        ylabel: str
            The label along the y-axis.  Defaults to 'energies [eV]'
        colors: sequence of str
            A sequence of one or two color specifications, depending on
            whether there is spin.
            Default: green if no spin, yellow and blue if spin is present.
        point_colors: ndarray
            An array of numbers of the shape (nspins, n_kpts, nbands) which
            are then mapped onto colors by the colormap (see ``cmap``).
            ``colors`` and ``point_colors`` are mutually exclusive
        label: str or list of str
            Label for the curves on the legend.  A string if one spin is
            present, a list of two strings if two spins are present.
            Default: If no spin is given, no legend is made; if spin is
            present default labels 'spin up' and 'spin down' are used, but
            can be suppressed by setting ``label=False``.
        loc: str
            Location of the legend.

        If ``point_colors`` is given, the following arguments can be specified.

        cmap:
            Only used if colors is an array of numbers.  A matplotlib
            colormap object, or a string naming a standard colormap.
            Default: The matplotlib default, typically 'viridis'.
        cmin, cmax: float
            Minimal and maximal values used for colormap translation.
            Default: -1.0 and 1.0
        colorbar: bool
            Whether to make a colorbar.
        clabel: str
            Label for the colorbar (default 's_z', set to None to suppress.
        cax: Axes
            Axes object used for plotting colorbar.  Default: split off a
            new one.
        sortcolors (bool or callable):
            Sort points so highest color values are in front.  If a callable is
            given, then it is called on the color values to determine the sort
            order.

        Any additional keyword arguments are passed directly to matplotlib's
        plot() or scatter() methods, depending on whether point_colors is
        given.
        """
        import matplotlib.pyplot as plt

        if colors is not None and point_colors is not None:
            raise ValueError("Don't give both 'color' and 'point_color'")

        if self.ax is None:
            ax = self.prepare_plot(ax, emin, emax, ylabel)

        e_skn = self.bs.energies
        nspins = len(e_skn)

        if point_colors is None:
            # Normal band structure plot
            if colors is None:
                if len(e_skn) == 1:
                    colors = 'g'
                else:
                    colors = 'yb'
            elif (len(colors) != nspins):
                raise ValueError(
                    "colors should be a sequence of {nspin} colors"
                )

            # Default values for label
            if label is None and nspins == 2:
                label = ['spin up', 'spin down']

            if label:
                if nspins == 1 and isinstance(label, str):
                    label = [label]
                elif len(label) != nspins:
                    raise ValueError(
                        f'label should be a list of {nspins} strings'
                    )

            for spin, e_kn in enumerate(e_skn):
                kwargs = dict(color=colors[spin])
                kwargs.update(plotkwargs)
                lbl = None   # Retain lbl=None if label=False
                if label:
                    lbl = label[spin]
                ax.plot(self.xcoords, e_kn[:, 0], label=lbl, **kwargs)

                for e_k in e_kn.T[1:]:
                    ax.plot(self.xcoords, e_k, **kwargs)
            show_legend = label is not None or nspins == 2

        else:
            # A color per datapoint.
            kwargs = dict(vmin=cmin, vmax=cmax, cmap=cmap, s=1)
            kwargs.update(plotkwargs)
            shape = e_skn.shape
            xcoords = np.zeros(shape)
            xcoords += self.xcoords[np.newaxis, :, np.newaxis]
            if sortcolors:
                if callable(sortcolors):
                    perm = sortcolors(point_colors).argsort(axis=None)
                else:
                    perm = point_colors.argsort(axis=None)
                e_skn = e_skn.ravel()[perm].reshape(shape)
                point_colors = point_colors.ravel()[perm].reshape(shape)
                xcoords = xcoords.ravel()[perm].reshape(shape)

            things = ax.scatter(xcoords, e_skn, c=point_colors, **kwargs)
            if colorbar:
                cbar = plt.colorbar(things, cax=cax)
                if clabel:
                    cbar.set_label(clabel)
            show_legend = False

        self.finish_plot(filename, show, loc, show_legend)

        return ax

    def prepare_plot(self, ax=None, emin=-10, emax=5, ylabel=None):
        import matplotlib.pyplot as plt
        if ax is None:
            ax = plt.figure().add_subplot(111)

        def pretty(kpt):
            if kpt == 'G':
                kpt = r'$\Gamma$'
            elif len(kpt) == 2:
                kpt = kpt[0] + '$_' + kpt[1] + '$'
            return kpt

        self.xcoords, label_xcoords, orig_labels = self.bs.get_labels()
        label_xcoords = list(label_xcoords)
        labels = [pretty(name) for name in orig_labels]

        i = 1
        while i < len(labels):
            if label_xcoords[i - 1] == label_xcoords[i]:
                labels[i - 1] = labels[i - 1] + ',' + labels[i]
                labels.pop(i)
                label_xcoords.pop(i)
            else:
                i += 1

        for x in label_xcoords[1:-1]:
            ax.axvline(x, color='0.5')

        ylabel = ylabel if ylabel is not None else 'energies [eV]'

        ax.set_xticks(label_xcoords)
        ax.set_xticklabels(labels)
        ax.set_ylabel(ylabel)
        ax.axhline(self.bs.reference, color='k', ls=':')
        ax.axis(xmin=0, xmax=self.xcoords[-1], ymin=emin, ymax=emax)
        self.ax = ax
        return ax

    def finish_plot(self, filename, show, loc, show_legend=False):
        import matplotlib.pyplot as plt

        if show_legend:
            leg = plt.legend(loc=loc)
            leg.get_frame().set_alpha(1)

        if filename:
            plt.savefig(filename)

        if show:
            plt.show()


@jsonable('bandstructure')
class BandStructure:
    """A band structure consists of an array of eigenvalues and a bandpath.

    BandStructure objects support JSON I/O.
    """

    def __init__(self, path, energies, reference=0.0):
        self._path = path
        self._energies = np.asarray(energies)
        assert self.energies.shape[0] in [1, 2]  # spins x kpts x bands
        assert self.energies.shape[1] == len(path.kpts)
        assert np.isscalar(reference)
        self._reference = reference

    @property
    def energies(self) -> np.ndarray:
        """The energies of this band structure.

        This is a numpy array of shape (nspins, nkpoints, nbands)."""
        return self._energies

    @property
    def path(self) -> 'ase.dft.kpoints.BandPath':
        """The :class:`~ase.dft.kpoints.BandPath` of this band structure."""
        return self._path

    @property
    def reference(self) -> float:
        """The reference energy.

        Semantics may vary; typically a Fermi energy or zero,
        depending on how the band structure was created."""
        return self._reference

    def subtract_reference(self) -> 'BandStructure':
        """Return new band structure with reference energy subtracted."""
        return BandStructure(self.path, self.energies - self.reference,
                             reference=0.0)

    def todict(self):
        return dict(path=self.path,
                    energies=self.energies,
                    reference=self.reference)

    def get_labels(self, eps=1e-5):
        """"See :func:`ase.dft.kpoints.labels_from_kpts`."""
        return self.path.get_linear_kpoint_axis(eps=eps)

    def plot(self, *args, **kwargs):
        """Plot this band structure."""
        bsp = BandStructurePlot(self)
        return bsp.plot(*args, **kwargs)

    def __repr__(self):
        return ('{}(path={!r}, energies=[{} values], reference={})'
                .format(self.__class__.__name__, self.path,
                        '{}x{}x{}'.format(*self.energies.shape),
                        self.reference))



from pathlib import Path

from ase.db.core import KeyDescription
from ase.db.row import row2dct
from ase.formula import Formula


class DatabaseProject:
    """Settings for web view of a database.

    For historical reasons called a "Project".
    """
    _ase_templates = Path('ase/db/templates')

    def __init__(self, name, title, *,
                 key_descriptions,
                 database,
                 default_columns):
        self.name = name
        self.title = title
        self.uid_key = 'id'

        # The templates loop over "key descriptions" when they want to
        # loop over keys.
        #
        # Therefore, any key without description will not be rendered.
        # Therefore, we need to make dummy key descriptions of everything
        # in the database, ensuring that all keys are visible.

        all_keys = database.get_all_key_names()

        key_descriptions = {
            **{key: KeyDescription(key) for key in all_keys},
            **key_descriptions}

        for key, value in key_descriptions.items():
            assert isinstance(key, str), type(key)
            assert isinstance(value, KeyDescription), type(value)

        self.key_descriptions = key_descriptions
        self.database = database
        self.default_columns = default_columns

    def get_search_template(self):
        return self._ase_templates / 'search.html'

    def get_row_template(self):
        return self._ase_templates / 'row.html'

    def get_table_template(self):
        return self._ase_templates / 'table.html'

    def handle_query(self, args) -> str:
        """Convert request args to ase.db query string."""
        return args['query']

    def row_to_dict(self, row):
        """Convert row to dict for use in html template."""
        dct = row2dct(row, self.key_descriptions)
        dct['formula'] = Formula(row.formula).convert('abc').format('html')
        return dct

    def uid_to_row(self, uid):
        return self.database.get(f'{self.uid_key}={uid}')

    @classmethod
    def dummyproject(cls, **kwargs):
        class DummyDatabase:
            def select(self, *args, **kwargs):
                return iter([])

            def get_all_key_names(self):
                return set()

        _kwargs = dict(
            name='test',
            title='test',
            key_descriptions={},
            database=DummyDatabase(),  # XXX
            default_columns=[])
        _kwargs.update(kwargs)
        return cls(**_kwargs)

    # If we make this a classmethod, and try to instantiate the class,
    # it would fail on subclasses.  So we use staticmethod
    @staticmethod
    def load_db_as_ase_project(name, database):
        from ase.db.core import get_key_descriptions
        from ase.db.table import all_columns

        return DatabaseProject(
            name=name,
            title=database.metadata.get('title', ''),
            key_descriptions=get_key_descriptions(),
            database=database,
            default_columns=all_columns)



from ase.build import bulk
from ase.calculators.emt import EMT
from ase.io import read
from ase.md import VelocityVerlet
from ase.units import fs
from ase.utils.forcecurve import force_curve


def test_forcecurve(testdir, plt):
    atoms = bulk('Au', cubic=True) * (2, 1, 1)
    atoms.calc = EMT()
    atoms.rattle(stdev=0.05)

    with VelocityVerlet(atoms, timestep=12.0 * fs,
                        trajectory='tmp.traj') as md:
        md.run(steps=10)
    images = read('tmp.traj', ':')
    force_curve(images)

    # import pylab as plt
    # plt.show()



import pytest

from ase.utils import xwopen

pytestmark = pytest.mark.usefixtures('testdir')


poem = 'Wer reitet so spät durch Nacht und Wind\n'.encode()
filename = 'poem.txt'


def test_xwopen():
    with xwopen(filename) as fd:
        fd.write(poem)

    assert fd.closed

    with open(filename, 'rb') as fd:
        assert fd.read() == poem


def test_xwopen_locked():
    with xwopen(filename) as fd:
        assert fd is not None
        with xwopen(filename) as fd2:
            assert fd2 is None


def test_xwopen_fail(tmp_path):
    with pytest.raises(OSError):
        with xwopen(tmp_path / 'does_not_exist/file'):
            pass



from ase.calculators.calculator import Calculator


def test_calculator_label():

    calc = Calculator()
    assert calc.directory == '.'
    assert calc.prefix is None
    assert calc.label is None

    calc.label = 'dir/pref'
    assert calc.directory == 'dir'
    assert calc.prefix == 'pref'
    assert calc.label == 'dir/pref'

    calc.label = 'dir2/'
    assert calc.directory == 'dir2'
    assert calc.prefix is None
    assert calc.label == 'dir2/'

    calc.label = 'hello'
    assert calc.directory == '.'
    assert calc.prefix == 'hello'
    assert calc.label == 'hello'

    calc.label = None
    assert calc.label is None
    assert calc.prefix is None
    assert calc.directory == '.'



import pytest

from ase import units
from ase.build import bulk
from ase.calculators.emt import EMT
from ase.io import Trajectory
from ase.md import VelocityVerlet
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution


@pytest.fixture(name="atoms")
def fixture_atoms():
    atoms = bulk("Au") * 2
    MaxwellBoltzmannDistribution(atoms, temperature_K=100.0)
    atoms.calc = EMT()
    return atoms


def test_md(atoms, testdir):
    def f():
        print(atoms.get_potential_energy(), atoms.get_total_energy())

    with VelocityVerlet(atoms, timestep=0.1) as md:
        md.attach(f)
        with Trajectory('Cu2.traj', 'w', atoms) as traj:
            md.attach(traj.write, interval=3)
            md.run(steps=20)

    with Trajectory('Cu2.traj', 'r') as traj:
        traj[-1]

    # Really?? No assertion at all?


@pytest.mark.parametrize("md_class", [VelocityVerlet])
def test_run_twice(md_class, atoms):
    """Test if `steps` increments `max_steps` when `run` is called twice."""
    steps = 5
    with md_class(atoms, timestep=1.0 * units.fs) as md:
        md.run(steps=steps)
        md.run(steps=steps)
    assert md.nsteps == 2 * steps
    assert md.max_steps == 2 * steps



import pytest

from ase.build import bulk

omx_par = {'definition_of_atomic_species': [['Al', 'Al8.0-p1', 'Al_CA13'],
                                            ['O', 'O6.0-p1', 'O_CA13']]}


calc = pytest.mark.calculator


# @calc('elk', rgkmax=5.0)
# Elk fails since calc.atoms is None after reading.
# Apparently this test did not run in CI before.
# Let's just disable it.
@calc('openmx', **omx_par)
def test_al(factory):
    name = factory.name
    # What on earth does kpts=1.0 mean?  Was failing, I changed it.  --askhl
    # Disabled GPAW since it was failing anyway. --askhl
    kpts = [2, 2, 2]
    calc = factory.calc(label=name, xc='LDA', kpts=kpts)
    al = bulk('AlO', crystalstructure='rocksalt', a=4.5)
    al.calc = calc
    e = al.get_potential_energy()
    calc.set(xc='PBE', kpts=kpts)
    epbe = al.get_potential_energy()
    print(e, epbe)
    calc = factory.calc(restart=name)
    print(calc.parameters, calc.results, calc.atoms)
    assert not calc.calculation_required(al, ['energy'])
    al = calc.get_atoms()
    print(al.get_potential_energy())
    label = 'dir/' + name + '-2'
    calc = factory.calc(label=label, atoms=al, xc='LDA', kpts=kpts)
    print(al.get_potential_energy())



import numpy as np
import pytest

from ase.build import bulk
from ase.calculators.emt import EMT
from ase.optimize import FIRE2


@pytest.mark.optimize
@pytest.mark.slow
def test_fire2():
    def system_setup():
        a = bulk('Au')
        a *= (2, 2, 2)
        a[0].x += 0.5
        a.calc = EMT()
        return a

    a = system_setup()

    fire_parameters = {'dt': 0.1,
                       'maxstep': 0.2,
                       'dtmax': 1.0,
                       'Nmin': 20,
                       'finc': 1.1,
                       'fdec': 0.5,
                       'astart': 0.25,
                       'fa': 0.99}

    opt = FIRE2(a,
                **fire_parameters)
    opt.run(fmax=0.001)
    e1 = a.get_potential_energy()
    n1 = opt.nsteps

    a = system_setup()

    reset_history = []

    def callback(a, r, e, e_last):
        reset_history.append([e - e_last])

    opt = FIRE2(a,
                use_abc=True,
                position_reset_callback=callback,
                **fire_parameters)
    opt.run(fmax=0.001)
    e2 = a.get_potential_energy()
    n2 = opt.nsteps

    assert e2 == pytest.approx(e1, abs=1e-6)
    assert n2 < n1
    assert all(np.array(reset_history) > 0)



import numpy as np

from ase import units
from ase.build import bulk
from ase.calculators.emt import EMT
from ase.md.langevin import Langevin
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution, Stationary


def test_langevin_com():
    """Check that the center of mass does not move during Langevin dynamics.

    In particular, test that this does not happen with atoms with different
    mass present (issue #1044).
    """
    # parameters
    size = 2
    T = 300
    dt = 0.01

    # setup
    atoms = bulk('CuAg', 'rocksalt', a=4.0).repeat(size)
    atoms.pbc = False
    atoms.calc = EMT()

    MaxwellBoltzmannDistribution(atoms, temperature_K=T)
    Stationary(atoms)

    mtot = atoms.get_momenta().sum(axis=0)
    print('initial momenta', mtot)
    print('initial forces', atoms.get_forces().sum(axis=0))

    # run NVT
    with Langevin(atoms, dt * units.fs, temperature_K=T, friction=0.02) as dyn:
        dyn.run(10)

    m2 = atoms.get_momenta().sum(axis=0)
    print('momenta', m2)
    print('forces', atoms.get_forces().sum(axis=0))
    print()

    assert np.linalg.norm(m2) < 1e-8



from unittest import mock

import numpy as np
import pytest

from ase.build import bulk
from ase.calculators.vasp.create_input import (
    GenerateVaspInput,
    _args_without_comment,
    _from_vasp_bool,
    _to_vasp_bool,
)


def dict_is_subset(d1, d2):
    """True if all the key-value pairs in dict 1 are in dict 2"""
    # Note, we are using direct comparison, so we should not compare
    # floats if any real computations are made, as that would be unsafe.
    # Cannot use pytest.approx here, because of of string comparison
    # not being available in python 3.6.
    return all(key in d2 and d1[key] == d2[key] for key in d1)


@pytest.fixture()
def rng():
    return np.random.RandomState(seed=42)


@pytest.fixture()
def nacl(rng):
    atoms = bulk('NaCl', crystalstructure='rocksalt', a=4.1,
                 cubic=True) * (3, 3, 3)
    rng.shuffle(atoms.symbols)  # Ensure symbols are mixed
    return atoms


@pytest.fixture()
def vaspinput_factory(nacl):
    """Factory for GenerateVaspInput class, which mocks the generation of
    pseudopotentials."""
    def _vaspinput_factory(atoms=None, **kwargs) -> GenerateVaspInput:
        if atoms is None:
            atoms = nacl
        mocker = mock.Mock()
        inputs = GenerateVaspInput()
        inputs.set(**kwargs)
        inputs._build_pp_list = mocker(  # type: ignore[method-assign]
            return_value=None
        )
        inputs.initialize(atoms)
        return inputs

    return _vaspinput_factory


def test_sorting(nacl, vaspinput_factory):
    """Test that the sorting/resorting scheme works"""
    vaspinput = vaspinput_factory(atoms=nacl)
    srt = vaspinput.sort
    resrt = vaspinput.resort
    atoms = nacl.copy()
    assert atoms[srt] != nacl
    assert atoms[resrt] != nacl
    assert atoms[srt][resrt] == nacl

    # Check the first and second half of the sorted atoms have the same symbols
    assert len(atoms) % 2 == 0  # We should have an even number of atoms
    atoms_sorted = atoms[srt]
    N = len(atoms) // 2
    seq1 = set(atoms_sorted.symbols[:N])
    seq2 = set(atoms_sorted.symbols[N:])
    assert len(seq1) == 1
    assert len(seq2) == 1
    # Check that we have two different symbols
    assert len(seq1.intersection(seq2)) == 0


@pytest.fixture(params=['random', 'ones', 'binaries'])
def magmoms_factory(rng, request):
    """Factory for generating various kinds of magnetic moments"""
    kind = request.param
    if kind == 'random':
        # Array of random
        func = rng.rand
    elif kind == 'ones':
        # Array of just 1's
        func = np.ones
    elif kind == 'binaries':
        # Array of 0's and 1's
        def rand_binary(x):
            return rng.randint(2, size=x)

        func = rand_binary
    else:
        raise ValueError(f'Unknown kind: {kind}')

    def _magmoms_factory(atoms):
        magmoms = func(len(atoms))
        assert len(magmoms) == len(atoms)
        return magmoms

    return _magmoms_factory


def read_magmom_from_file(filename) -> np.ndarray:
    """Helper function to parse the magnetic moments from an INCAR file"""
    found = False
    with open(filename) as file:
        for line in file:
            # format "MAGMOM = n1*val1 n2*val2 ..."
            if 'MAGMOM = ' in line:
                found = True
                parts = line.strip().split()[2:]
                new_magmom = []
                for part in parts:
                    n, val = part.split('*')
                    # Add "val" to magmom "n" times
                    new_magmom += int(n) * [float(val)]
                break
    assert found
    return np.array(new_magmom)


@pytest.fixture()
def assert_magmom_equal_to_incar_value():
    """Fixture to compare a pre-made magmom array to the value
    a GenerateVaspInput.write_incar object writes to a file"""
    def _assert_magmom_equal_to_incar_value(atoms, expected_magmom, vaspinput):
        assert len(atoms) == len(expected_magmom)
        vaspinput.write_incar(atoms)
        new_magmom = read_magmom_from_file('INCAR')
        assert len(new_magmom) == len(expected_magmom)
        srt = vaspinput.sort
        resort = vaspinput.resort
        # We round to 4 digits
        assert np.allclose(expected_magmom, new_magmom[resort], atol=1e-3)
        assert np.allclose(np.array(expected_magmom)[srt],
                           new_magmom,
                           atol=1e-3)

    return _assert_magmom_equal_to_incar_value


@pytest.mark.parametrize('list_func', [list, tuple, np.array])
def test_write_magmom(magmoms_factory, list_func, nacl, vaspinput_factory,
                      assert_magmom_equal_to_incar_value, testdir):
    """Test writing magnetic moments to INCAR, and ensure we can do it
    passing different types of sequences"""
    magmom = magmoms_factory(nacl)

    vaspinput = vaspinput_factory(atoms=nacl, magmom=magmom, ispin=2)
    assert vaspinput.spinpol
    assert_magmom_equal_to_incar_value(nacl, magmom, vaspinput)


def test_atoms_with_initial_magmoms(magmoms_factory, nacl, vaspinput_factory,
                                    assert_magmom_equal_to_incar_value,
                                    testdir):
    """Test passing atoms with initial magnetic moments"""
    magmom = magmoms_factory(nacl)
    assert len(magmom) == len(nacl)
    nacl.set_initial_magnetic_moments(magmom)
    vaspinput = vaspinput_factory(atoms=nacl)
    assert vaspinput.spinpol
    assert_magmom_equal_to_incar_value(nacl, magmom, vaspinput)


def test_vasp_from_bool():
    for s in ('T', '.true.'):
        assert _from_vasp_bool(s) is True
    for s in ('f', '.False.'):
        assert _from_vasp_bool(s) is False
    with pytest.raises(ValueError):
        _from_vasp_bool('yes')
    with pytest.raises(AssertionError):
        _from_vasp_bool(True)


def test_vasp_to_bool():
    for x in ('T', '.true.', True):
        assert _to_vasp_bool(x) == '.TRUE.'
    for x in ('f', '.FALSE.', False):
        assert _to_vasp_bool(x) == '.FALSE.'

    with pytest.raises(ValueError):
        _to_vasp_bool('yes')
    with pytest.raises(AssertionError):
        _to_vasp_bool(1)


@pytest.mark.parametrize('args, expected_len',
                         [(['a', 'b', '#', 'c'], 2),
                          (['a', 'b', '!', 'c', '#', 'd'], 2),
                          (['#', 'a', 'b', '!', 'c', '#', 'd'], 0)])
def test_vasp_args_without_comment(args, expected_len):
    """Test comment splitting logic"""
    clean_args = _args_without_comment(args)
    assert len(clean_args) == expected_len


def test_vasp_xc(vaspinput_factory):
    """
    Run some tests to ensure that the xc setting in the VASP calculator
    works.
    """

    calc_vdw = vaspinput_factory(xc='optb86b-vdw')

    assert dict_is_subset({
        'param1': 0.1234,
        'param2': 1.0
    }, calc_vdw.float_params)
    assert calc_vdw.bool_params['luse_vdw'] is True

    calc_hse = vaspinput_factory(xc='hse06',
                                 hfscreen=0.1,
                                 gga='RE',
                                 encut=400,
                                 sigma=0.5)

    assert dict_is_subset({
        'hfscreen': 0.1,
        'encut': 400,
        'sigma': 0.5
    }, calc_hse.float_params)
    assert calc_hse.bool_params['lhfcalc'] is True
    assert dict_is_subset({'gga': 'RE'}, calc_hse.string_params)

    with pytest.warns(FutureWarning):
        calc_pw91 = vaspinput_factory(xc='pw91',
                                      kpts=(2, 2, 2),
                                      gamma=True,
                                      lreal='Auto')
        assert dict_is_subset(
            {
                'pp': 'PW91',
                'kpts': (2, 2, 2),
                'gamma': True,
                'reciprocal': False
            }, calc_pw91.input_params)


def test_ichain(vaspinput_factory):

    with pytest.warns(UserWarning):
        calc_warn = vaspinput_factory(ichain=1, ediffg=-0.01)
        calc_warn.write_incar(nacl)
        calc_warn.read_incar('INCAR')
        assert calc_warn.int_params['iopt'] == 1
        assert calc_warn.exp_params['ediffg'] == -0.01
        assert calc_warn.int_params['ibrion'] == 3
        assert calc_warn.float_params['potim'] == 0.0

    with pytest.raises(RuntimeError):
        calc_wrong = vaspinput_factory(ichain=1, ediffg=0.0001, iopt=1)
        calc_wrong.write_incar(nacl)
        calc_wrong.read_incar('INCAR')
        assert calc_wrong.int_params['iopt'] == 1

    calc = vaspinput_factory(ichain=1,
                             ediffg=-0.01,
                             iopt=1,
                             potim=0.0,
                             ibrion=3)
    calc.write_incar(nacl)
    calc.read_incar('INCAR')
    assert calc.int_params['iopt'] == 1
    assert calc.exp_params['ediffg'] == -0.01
    assert calc.int_params['ibrion'] == 3
    assert calc.float_params['potim'] == 0.0


def test_non_registered_keys(vaspinput_factory) -> None:
    """Test if non-registered INCAR keys can be written and read.

    Here the SCAN meta-GGA functional via LIBXC is tested.

    https://www.vasp.at/wiki/index.php/LIBXC1#Examples_of_INCAR

    """
    calc = vaspinput_factory(metagga='LIBXC')

    # Be sure that `libxc1` and `libxc2` are not in the registered INCAR keys.
    assert 'libxc1' not in calc.string_params
    assert 'libxc2' not in calc.int_params

    calc.set(libxc1='MGGA_X_SCAN')  # or 263
    calc.set(libxc2=267)  # or "MGGA_C_SCAN"

    calc.write_incar(nacl)
    calc.read_incar('INCAR')

    assert calc.string_params['libxc1'] == 'MGGA_X_SCAN'
    assert calc.int_params['libxc2'] == 267


def test_bool(tmp_path, vaspinput_factory):
    """Test that INCAR parser behaves similarly to Vasp, which uses
    default fortran 'read' parsing
    """

    for bool_str in ['t', 'T', 'true', 'TRUE', 'TrUe', '.true.', '.T', 'tbob']:
        with open(tmp_path / 'INCAR', 'w') as fout:
            fout.write('ENCUT = 100\n')
            fout.write(f'LCHARG = {bool_str}\n')
        calc = vaspinput_factory(encut=100)
        calc.read_incar(tmp_path / 'INCAR')
        assert calc.bool_params['lcharg']

    for bool_str in ['f', 'F', 'false', 'FALSE', 'FaLSe', '.false.', '.F',
                     'fbob']:
        with open(tmp_path / 'INCAR', 'w') as fout:
            fout.write('ENCUT = 100\n')
            fout.write(f'LCHARG = {bool_str}\n')
        calc = vaspinput_factory(encut=100)
        calc.read_incar(tmp_path / 'INCAR')
        assert not calc.bool_params['lcharg']

    for bool_str in ['x', '..true.', '1']:
        with open(tmp_path / 'INCAR', 'w') as fout:
            fout.write('ENCUT = 100\n')
            fout.write(f'LCHARG = {bool_str}\n')
        calc = vaspinput_factory(encut=100)
        with pytest.raises(ValueError):
            calc.read_incar(tmp_path / 'INCAR')



import numpy as np

from ase.build import bulk
from ase.md import Langevin
from ase.md.fix import FixRotation
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution, Stationary
from ase.units import fs
from ase.utils import seterr


def check_inertia(atoms):
    m, v = atoms.get_moments_of_inertia(vectors=True)
    print("Moments of inertia:")
    print(m)
    # There should be one vector in the z-direction
    n = 0
    delta = 1e-2
    for a in v:
        if (abs(a[0]) < delta
            and abs(a[1]) < delta
                and abs(abs(a[2]) - 1.0) < delta):

            print("Vector along z:", a)
            n += 1
        else:
            print("Vector not along z:", a)
    assert n == 1


def test_fixrotation_asap(asap3):
    rng = np.random.RandomState(123)

    with seterr(all='raise'):
        atoms = bulk('Au', cubic=True).repeat((3, 3, 10))
        atoms.pbc = False
        atoms.center(vacuum=5.0 + np.max(atoms.cell) / 2)
        print(atoms)
        atoms.calc = asap3.EMT()
        MaxwellBoltzmannDistribution(atoms, temperature_K=300, force_temp=True,
                                     rng=rng)
        Stationary(atoms)
        check_inertia(atoms)
        com = atoms.get_center_of_mass()
        with Langevin(
                atoms,
                timestep=20 * fs,
                temperature_K=300,
                friction=1e-3,
                logfile='-',
                loginterval=500,
                rng=rng
        ) as md:
            fx = FixRotation(atoms)
            md.attach(fx)
            md.run(steps=1000)
        check_inertia(atoms)
        # Test for issue #977 (it is free to do so here).
        delta = np.linalg.norm(atoms.get_center_of_mass() - com)
        print("Change in center of mass:", delta)
        assert delta < 1e-9



import re

_pattern_test_data = []


def _define_pattern(pattern, example, *args):
    """Compile a regex and store an example pattern for testing."""
    regex = re.compile(pattern, *args)
    _pattern_test_data.append((regex, example))
    return regex



import pytest

from ase.build import bulk
from ase.visualize import x3d


def test_x3d():
    pytest.importorskip('IPython')
    from IPython.display import HTML

    atoms = bulk('Cu', cubic=True)
    my_obj = x3d.view_x3d(atoms)
    assert isinstance(my_obj, HTML)



"""Prism"""
import warnings
from typing import Sequence

import numpy as np

from ase.geometry import wrap_positions


def calc_box_parameters(cell: np.ndarray) -> np.ndarray:
    """Calculate box parameters

    https://docs.lammps.org/Howto_triclinic.html
    """
    ax = np.sqrt(cell[0] @ cell[0])
    bx = cell[0] @ cell[1] / ax
    by = np.sqrt(cell[1] @ cell[1] - bx ** 2)
    cx = cell[0] @ cell[2] / ax
    cy = (cell[1] @ cell[2] - bx * cx) / by
    cz = np.sqrt(cell[2] @ cell[2] - cx ** 2 - cy ** 2)
    return np.array((ax, by, cz, bx, cx, cy))


def calc_rotated_cell(cell: np.ndarray) -> np.ndarray:
    """Calculate rotated cell in LAMMPS coordinates

    Parameters
    ----------
    cell : np.ndarray
        Cell to be rotated.

    Returns
    -------
    rotated_cell : np.ndarray
        Rotated cell represented by a lower triangular matrix.
    """
    ax, by, cz, bx, cx, cy = calc_box_parameters(cell)
    return np.array(((ax, 0.0, 0.0), (bx, by, 0.0), (cx, cy, cz)))


def calc_reduced_cell(cell: np.ndarray, pbc: Sequence[bool]) -> np.ndarray:
    """Calculate LAMMPS cell with short lattice basis vectors

    The lengths of the second and the third lattice basis vectors, b and c, are
    shortened with keeping the same periodicity of the system. b is modified by
    adding multiple a vectors, and c is modified by adding first multiple b
    vectors and then multiple a vectors.

    Parameters
    ----------
    cell : np.ndarray
        Cell to be reduced. This must be already a lower triangular matrix.
    pbc : Sequence[bool]
        True if the system is periodic along the corresponding direction.

    Returns
    -------
    reduced_cell : np.ndarray
        Reduced cell. `xx`, `yy`, `zz` are the same as the original cell,
        and `abs(xy) <= xx`, `abs(xz) <= xx`, `abs(yz) <= yy`.
    """
    # cell 1 (reduced) <- cell 2 (original)
    # o-----------------------------/==o-----------------------------/--o
    #  \                        /--/    \                        /--/
    #   \                   /--/         \                   /--/
    #    \         1    /--/              \   2          /--/
    #     \         /--/                   \         /--/
    #      \    /--/                        \    /--/
    #       o==/-----------------------------o--/
    reduced_cell = cell.copy()
    # Order in which off-diagonal elements are checked for strong tilt
    # yz is updated before xz so that the latter does not affect the former
    flip_order = ((1, 0), (2, 1), (2, 0))
    for i, j in flip_order:
        if not pbc[j]:
            continue
        ratio = reduced_cell[i, j] / reduced_cell[j, j]
        if abs(ratio) > 0.5:
            reduced_cell[i] -= reduced_cell[j] * np.round(ratio)
    return reduced_cell


class Prism:
    """The representation of the unit cell in LAMMPS

    The main purpose of the prism-object is to create suitable string
    representations of prism limits and atom positions within the prism.

    Parameters
    ----------
    cell : np.ndarray
        Cell in ASE coordinate system.
    pbc : one or three bool
        Periodic boundary conditions flags.
    reduce_cell : bool
        If True, the LAMMPS cell is reduced for short lattice basis vectors.
        The atomic positions are always wraped into the reduced cell,
        regardress of `wrap` in `vector_to_lammps` and `vector_to_ase`.
    tolerance : float
        Precision for skewness test.

    Methods
    -------
    vector_to_lammps
        Rotate vectors from ASE to LAMMPS coordinates.
        Positions can be further wrapped into the LAMMPS cell by `wrap=True`.

    vector_to_ase
        Rotate vectors from LAMMPS to ASE coordinates.
        Positions can be further wrapped into the LAMMPS cell by `wrap=True`.

    Notes
    -----
    LAMMPS prefers triangular matrixes without a strong tilt.
    Therefore the 'Prism'-object contains three coordinate systems:

    - ase_cell (the simulated system in the ASE coordination system)
    - lammps_tilt (ase-cell rotated to be an lower triangular matrix)
    - lammps_cell (same volume as tilted cell, but reduce edge length)

    The translation between 'ase_cell' and 'lammps_tilt' is done with a
    rotation matrix 'rot_mat'. (Mathematically this is a QR decomposition.)

    The transformation between 'lammps_tilt' and 'lammps_cell' is done by
    changing the off-diagonal elements.

    Depending on the option `reduce`, vectors in ASE coordinates are
    transformed either `lammps_tilt` or `lammps_cell`.

    The vector conversion can fail as depending on the simulation run LAMMPS
    might have changed the simulation box significantly. This is for example a
    problem with hexagonal cells. LAMMPS might also wrap atoms across periodic
    boundaries, which can lead to problems for example NEB calculations.
    """

    # !TODO: derive tolerance from cell-dimensions
    def __init__(
        self,
        cell: np.ndarray,
        pbc: bool = True,
        reduce_cell: bool = False,
        tolerance: float = 1.0e-8,
    ):
        #    rot_mat * lammps_tilt^T = ase_cell^T
        # => lammps_tilt * rot_mat^T = ase_cell
        # => lammps_tilt             = ase_cell * rot_mat
        # LAMMPS requires positive diagonal elements of the triangular matrix.
        # The diagonals of `lammps_tilt` are always positive by construction.
        self.lammps_tilt = calc_rotated_cell(cell)
        self.rot_mat = np.linalg.solve(self.lammps_tilt, cell).T
        self.ase_cell = cell
        self.tolerance = tolerance
        self.pbc = tuple(np.zeros(3, bool) + pbc)
        self.lammps_cell = calc_reduced_cell(self.lammps_tilt, self.pbc)
        self.is_reduced = reduce_cell

    @property
    def cell(self) -> np.ndarray:
        return self.lammps_cell if self.is_reduced else self.lammps_tilt

    def get_lammps_prism(self) -> np.ndarray:
        """Return box parameters of the rotated cell in LAMMPS coordinates

        Returns
        -------
        np.ndarray
            xhi - xlo, yhi - ylo, zhi - zlo, xy, xz, yz
        """
        return self.cell[(0, 1, 2, 1, 2, 2), (0, 1, 2, 0, 0, 1)]

    def update_cell(self, lammps_cell: np.ndarray) -> np.ndarray:
        """Rotate new LAMMPS cell into ASE coordinate system

        Parameters
        ----------
        lammps_cell : np.ndarray
            New Cell in LAMMPS coordinates received after executing LAMMPS

        Returns
        -------
        np.ndarray
            New cell in ASE coordinates
        """
        # Transformation: integer matrix
        # lammps_cell * transformation = lammps_tilt
        transformation = np.linalg.solve(self.lammps_cell, self.lammps_tilt)

        if self.is_reduced:
            self.lammps_cell = lammps_cell
            self.lammps_tilt = lammps_cell @ transformation
        else:
            self.lammps_tilt = lammps_cell
            self.lammps_cell = calc_reduced_cell(self.lammps_tilt, self.pbc)

        # try to detect potential flips in lammps
        # (lammps minimizes the cell-vector lengths)
        new_ase_cell = self.lammps_tilt @ self.rot_mat.T
        # assuming the cell changes are mostly isotropic
        new_vol = np.linalg.det(new_ase_cell)
        old_vol = np.linalg.det(self.ase_cell)
        test_residual = self.ase_cell.copy()
        test_residual *= (new_vol / old_vol) ** (1.0 / 3.0)
        test_residual -= new_ase_cell
        if any(
                np.linalg.norm(test_residual, axis=1)
                > 0.5 * np.linalg.norm(self.ase_cell, axis=1)
        ):
            warnings.warn(
                "Significant simulation cell changes from LAMMPS detected. "
                "Backtransformation to ASE might fail!"
            )
        return new_ase_cell

    def vector_to_lammps(
        self,
        vec: np.ndarray,
        wrap: bool = False,
    ) -> np.ndarray:
        """Rotate vectors from ASE to LAMMPS coordinates

        Parameters
        ----------
        vec : np.ndarray
            Vectors in ASE coordinates to be rotated into LAMMPS coordinates
        wrap : bool
            If True, the vectors are wrapped into the cell

        Returns
        -------
        np.array
            Vectors in LAMMPS coordinates
        """
        # !TODO: right eps-limit
        # lammps might not like atoms outside the cell
        if wrap or self.is_reduced:
            return wrap_positions(
                vec @ self.rot_mat,
                cell=self.cell,
                pbc=self.pbc,
                eps=1e-18,
            )
        return vec @ self.rot_mat

    def vector_to_ase(
        self,
        vec: np.ndarray,
        wrap: bool = False,
    ) -> np.ndarray:
        """Rotate vectors from LAMMPS to ASE coordinates

        Parameters
        ----------
        vec : np.ndarray
            Vectors in LAMMPS coordinates to be rotated into ASE coordinates
        wrap : bool
            If True, the vectors are wrapped into the cell

        Returns
        -------
        np.ndarray
            Vectors in ASE coordinates
        """
        if wrap or self.is_reduced:
            # fractional in `lammps_tilt` (the same shape as ASE cell)
            fractional = np.linalg.solve(self.lammps_tilt.T, vec.T).T
            # wrap into 0 to 1 for periodic directions
            fractional -= np.floor(fractional) * self.pbc
            # Cartesian coordinates wrapped into `lammps_tilt`
            vec = fractional @ self.lammps_tilt
        # rotate back to the ASE cell
        return vec @ self.rot_mat.T

    def tensor2_to_ase(self, tensor: np.ndarray) -> np.ndarray:
        """Rotate a second order tensor from LAMMPS to ASE coordinates

        Parameters
        ----------
        tensor : np.ndarray
            Tensor in LAMMPS coordinates to be rotated into ASE coordinates

        Returns
        -------
        np.ndarray
            Tensor in ASE coordinates
        """
        return self.rot_mat @ tensor @ self.rot_mat.T

    def is_skewed(self) -> bool:
        """Test if the lammps cell is skewed, i.e., monoclinic or triclinic.

        Returns
        -------
        bool
            True if the lammps cell is skewed.
        """
        cell_sq = self.cell ** 2
        on_diag = np.sum(np.diag(cell_sq))
        off_diag = np.sum(np.tril(cell_sq, -1))
        return off_diag / on_diag > self.tolerance
